"""Utility functions for logging operations."""

__author__ = 'Jeroen Van Der Donckt'

import logging
import warnings
import pandas as pd

from pathlib import Path
from typing import Union


def delete_logging_handlers(logger: logging.Logger):
    """Delete all logging handlers that are not stream-handlers.

    Parameters
    ----------
    logger : logging.Logger
        The logger.
    
    """
    if len(logger.handlers) > 1:
        logger.handlers = [
            h for h in logger.handlers if type(h) == logging.StreamHandler
        ]
    assert len(logger.handlers) == 1, "Multiple logging StreamHandlers present!!"

def add_logging_handler(logger: logging.Logger, logging_file_path: Union[str, Path]):
    """Add a logging file-handler to the logger.

    Parameters
    ----------
    logger : logging.Logger
        The logger.
     logging_file_path : Union[str, Path]
        The file path for the file handler.
    
    """
    if not isinstance(logging_file_path, Path):
        logging_file_path = Path(logging_file_path)
    if logging_file_path.exists():
        warnings.warn(
            f"Logging file ({logging_file_path}) already exists. "
            f"This file will be overwritten!",
            RuntimeWarning,
        )
        # Clear the file
        #  -> because same FileHandler is used when calling this method twice
        open(logging_file_path, "w").close()
    f_handler = logging.FileHandler(logging_file_path, mode="w")
    f_handler.setFormatter(
        logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
    )
    f_handler.setLevel(logging.INFO)
    logger.addHandler(f_handler)


def logging_file_to_df(logging_file_path: str) -> pd.DataFrame:
    """Parse the logged messages into a dataframe.

    Parameters
    ----------
    logging_file_path: str
        The file path where the logged messages are stored.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the log_time, name, log_level and log message.

    """
    column_names = ["log_time", "name", "log_level", "message"]
    data = {col: [] for col in column_names}
    with open(logging_file_path, "r") as f:
        for line in f:
            line = line.split(" - ")
            for idx, col in enumerate(column_names):
                data[col].append(line[idx].strip())
    df = pd.DataFrame(data)
    df["log_time"] = pd.to_datetime(df["log_time"])
    return df



"""Pipeline class that wraps sklearn.pipeline."""

__author__ = "Jeroen Van Der Donckt"

import imblearn
import pandas as pd
import numpy as np

from sklearn.utils.validation import check_memory

from sklearn.utils.metaestimators import if_delegate_has_method

from .SegmenterMixin import SegmenterMixin
from ..processing import SKSeriesPipeline
from ..features import SKFeatureCollection
from .dataframe_oprator import to_dataframe_operator

def make_pipeline(steps, memory=None, verbose=False):
        """Create a pipeline."""        
        # Wrap the steps of the pipeline into a DataFrameOperator if necessary
        def wrap_step(step):
            # It is not necessary to wrap SKSeriesPipeline or SKFeatureCollection as
            # these transformers return a dataframe by default :)
            if (
                isinstance(step[1], SKSeriesPipeline) or 
                isinstance(step[1], SKFeatureCollection)
            ):
                return step
            return (step[0], to_dataframe_operator(step[1]))
        steps = [wrap_step(step) for step in steps]
        return Pipeline(steps, memory=memory, verbose=verbose)


class Pipeline(imblearn.pipeline.Pipeline):

    def _apply_transformers(self, X: pd.DataFrame, y: np.ndarray = None):
        Xt = X
        yt = y
        for _, transform in self.steps[:-1]:
            if transform is None:
                continue
            if hasattr(transform, "segment"):
                segmenter = transform
                Xt, yt = segmenter.segment(Xt, yt)
            elif hasattr(transform, "fit_sample"):
                pass
            else:
               Xt = transform.transform(Xt)
        return Xt, yt


    def _fit(self, X, y=None, **fit_params):
        # shallow copy of steps - this should really be steps_
        self.steps = list(self.steps)
        self._validate_steps()
        # Setup the memory
        memory = check_memory(self.memory)

        fit_transform_one_cached = memory.cache(_fit_transform_one)
        fit_sample_one_cached = memory.cache(_fit_sample_one)

        for (step_idx, name, transformer) in self._iter(
            with_final=False, filter_passthrough=False
        ):
            if transformer is None or transformer == "passthrough":
                with _print_elapsed_time("Pipeline", self._log_message(step_idx)):
                    continue

            if hasattr(memory, "location"):
                # joblib >= 0.12
                if memory.location is None:
                    # we do not clone when caching is disabled to
                    # preserve backward compatibility
                    cloned_transformer = transformer
                else:
                    cloned_transformer = clone(transformer)
            elif hasattr(memory, "cachedir"):
                # joblib < 0.11
                if memory.cachedir is None:
                    # we do not clone when caching is disabled to
                    # preserve backward compatibility
                    cloned_transformer = transformer
                else:
                    cloned_transformer = clone(transformer)
            else:
                cloned_transformer = clone(transformer)
            # Fit or load from cache the current transformer
            X, fitted_transformer = fit_transform_one_cached(
                cloned_transformer,
                X,
                y,
                None,
                message_clsname="Pipeline",
                message=self._log_message(step_idx),
                **fit_params_steps[name],
            )
            # Replace the transformer of the step with the fitted
            # transformer. This is necessary when loading the transformer
            # from the cache.
            self.steps[step_idx] = (name, fitted_transformer)
        return X

    def fit(self, X, y=None, **fit_params):
        """Fit the model
        Fit all the transforms one after the other and transform the
        data, then fit the transformed data using the final estimator.
        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of the
            pipeline.
        y : iterable, default=None
            Training targets. Must fulfill label requirements for all steps of
            the pipeline.
        **fit_params : dict of string -> object
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.
        Returns
        -------
        self : Pipeline
            This estimator
        """
        fit_params_steps = self._check_fit_params(**fit_params)
        Xt = self._fit(X, y, **fit_params_steps)
        with _print_elapsed_time("Pipeline", self._log_message(len(self.steps) - 1)):
            if self._final_estimator != "passthrough":
                fit_params_last_step = fit_params_steps[self.steps[-1][0]]
                self._final_estimator.fit(Xt, y, **fit_params_last_step)

        return self

        # fit_params_steps = dict((name, {}) for name, step in self.steps
        #                         if step is not None)
        # for pname, pval in six.iteritems(fit_params):
        #     step, param = pname.split('__', 1)
        #     fit_params_steps[step][param] = pval
        # Xt = X
        # yt = y
        # for step_idx, (name, transformer) in enumerate(self.steps[:-1]):
        #     if transformer is None:
        #         pass
        #     else:
        #         if memory.cachedir is None:
        #             # we do not clone when caching is disabled to preserve
        #             # backward compatibility
        #             cloned_transformer = transformer
        #         else:
        #             cloned_transformer = clone(transformer)
        #         # Fit or load from cache the current transfomer
        #         if (hasattr(cloned_transformer, "transform") or
        #                 hasattr(cloned_transformer, "fit_transform")):
        #             Xt, fitted_transformer = fit_transform_one_cached(
        #                 cloned_transformer, None, Xt, yt,
        #                 **fit_params_steps[name])
        #         elif hasattr(cloned_transformer, "sample"):
        #             Xt, yt, fitted_transformer = fit_sample_one_cached(
        #                 cloned_transformer, Xt, yt,
        #                 **fit_params_steps[name])
        #         # Replace the transformer of the step with the fitted
        #         # transformer. This is necessary when loading the transformer
        #         # from the cache.
        #         self.steps[step_idx] = (name, fitted_transformer)
        # if self._final_estimator is None:
        #     return Xt, yt, {}
        # return Xt, yt, fit_params_steps[self.steps[-1][0]]

    @if_delegate_has_method(delegate='_final_estimator')
    def predict(self, X):
        """Apply transformers/samplers to the data, and predict with the final
        estimator
        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.
        Returns
        -------
        y_pred : array-like
        """
        Xt, _ = self._apply_transformers(X)
        return self.steps[-1][-1].predict(Xt)

    @if_delegate_has_method(delegate='_final_estimator')
    def predict_proba(self, X):
        """Apply transformers/samplers, and predict_proba of the final
        estimator
        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.
        Returns
        -------
        y_proba : array-like, shape = [n_samples, n_classes]
        """
        Xt, _ = self._apply_transformers(X)
        return self.steps[-1][-1].predict_proba(Xt)

    @if_delegate_has_method(delegate='_final_estimator')
    def score(self, X, y=None, sample_weight=None):
        """Apply transformers/samplers, and score with the final estimator
        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.
        y : iterable, default=None
            Targets used for scoring. Must fulfill label requirements for all
            steps of the pipeline.
        sample_weight : array-like, default=None
            If not None, this argument is passed as ``sample_weight`` keyword
            argument to the ``score`` method of the final estimator.
        Returns
        -------
        score : float
        """
        Xt, yt = self._apply_transformers(X, y)
        score_params = {}
        if sample_weight is not None:
            score_params['sample_weight'] = sample_weight
        return self.steps[-1][-1].score(Xt, yt, **score_params)


# RECHTSTREEKS GEKOPIEERD
def _fit_transform_one(
    transformer, X, y, weight, message_clsname="", message=None, **fit_params
):
    """
    Fits ``transformer`` to ``X`` and ``y``. The transformed result is returned
    with the fitted transformer. If ``weight`` is not ``None``, the result will
    be multiplied by ``weight``.
    """
    with _print_elapsed_time(message_clsname, message):
        if hasattr(transformer, "fit_transform"):
            res = transformer.fit_transform(X, y, **fit_params)
        else:
            res = transformer.fit(X, y, **fit_params).transform(X)

    if weight is None:
        return res, transformer
    return res * weight, transformer

def _fit_sample_one(sampler, X, y, **fit_params):
    X_res, y_res = sampler.fit_sample(X, y, **fit_params)

    return X_res, y_res, sampler

    # def _validate_steps(self):
    #     names, estimators = zip(*self.steps)

    #     # validate names
    #     self._validate_names(names)

    #     # validate estimators
    #     transformers = estimators[:-1]
    #     estimator = estimators[-1]

    #     for t in transformers:
    #         if t is None:
    #             continue
    #         if (not (hasattr(t, "fit") or
    #                  hasattr(t, "fit_transform") or
    #                  hasattr(t, "fit_sample")) or
    #             not (hasattr(t, "transform") or
    #                  hasattr(t, "sample"))):
    #             raise TypeError(
    #                 "All intermediate steps of the chain should "
    #                 "be estimators that implement fit and transform or sample "
    #                 "(but not both) '%s' (type %s) doesn't)" % (t, type(t)))

    #         # if ((hasattr(t, "fit_sample") and
    #         #      hasattr(t, "fit_transform")) or
    #         #     (hasattr(t, "sample") and
    #         #      hasattr(t, "transform"))):
    #         #     raise TypeError(
    #         #         "All intermediate steps of the chain should "
    #         #         "be estimators that implement fit and transform or sample."
    #         #         " '%s' implements both)" % (t))

    #         if isinstance(t, Pipeline):
    #             raise TypeError(
    #                 "All intermediate steps of the chain should not be"
    #                 " Pipelines")

    #     # We allow last estimator to be None as an identity transformation
    #     if estimator is not None and not hasattr(estimator, "fit"):
    #         raise TypeError("Last step of Pipeline should implement fit. "
    #                         "'%s' (type %s) doesn't")


# class Pipeline(Pipeline):

#     def __init__(self, steps, memory=None, verbose=False):
#         """Create a pipeline."""        
#         # Wrap the steps of the pipeline into a DataFrameOperator if necessary
#         def wrap_step(step):
#             # It is not necessary to wrap SKSeriesPipeline or SKFeatureCollection as
#             # these transformers return a dataframe by default :)
#             if (
#                 isinstance(step[1], SKSeriesPipeline) or 
#                 isinstance(step[1], SKFeatureCollection)
#             ):
#                 return step
#             return (step[0], to_dataframe_operator(step[1]))
#         steps = [wrap_step(step) for step in steps]
#         super().__init__(steps, memory=memory, verbose=verbose)



"""NumpyFuncWrapper class for object-oriented representation of a function."""

__author__ = "Jonas Van Der Donckt, Jeroen Van Der Donckt, Emiel Deprost"

from typing import Callable, List, Union, Any

import numpy as np

from .. import __pdoc__
from ..utils.classes import FrozenClass

__pdoc__['NumpyFuncWrapper.__call__'] = True


class NumpyFuncWrapper(FrozenClass):  # TODO: waarom niet gewoon FuncWrapper?
    """Numpy function wrapper.

    A Numpy function wrapper which takes a numpy array as input and returns a numpy
    array. It also defines the names of the function outputs.

    Parameters
    ----------
    func : Callable
        The wrapped function.
    output_names : Union[List[str], str], optional
        The name of the outputs of the function, by default None.

    Raises
    ------
    TypeError
        Raised when the `output_names` cannot be set.

    """

    def __init__(
        self, func: Callable, output_names: Union[List[str], str] = None, **kwargs
    ):
        """Create NumpyFuncWrapper instance."""
        self.func = func
        self.kwargs: dict = kwargs

        if isinstance(output_names, list):
            self.output_names = output_names
        elif isinstance(output_names, str):
            self.output_names = [output_names]
        elif not output_names:
            self.output_names = [self.func.__name__]
        else:
            raise TypeError(f"`output_names` is unexpected type {type(output_names)}")

        self._freeze()

    def __repr__(self) -> str:
        """Return repr string."""
        return (
            f"{self.__class__.__name__}({self.func.__name__}, {self.output_names},"
            f" {self.kwargs})"
        )

    def __call__(self, *series: np.ndarray) -> Any:
        """Call wrapped function with passed data.

        Parameters
        ---------
        *series : np.ndarray
            The (multiple) input series for the function.

        Returns
        -------
        Any
            The function its output for the passed series.

        """
        return self.func(*series, **self.kwargs)



# -*- coding: utf-8 -*-
"""Utilities for chunking time-series data before feeding it to the operators.
"""

__author__ = 'Jonas Van Der Donckt'

from .chunking import chunk_data

__all__ = [
    "chunk_data"
]



"""Feature extraction submodule.

.. include:: ../../docs/pdoc_include/features.md

"""

__author__ = "Jonas Van Der Donckt, Jeroen Van Der Donckt, Emiel Deprost"

from .. import __pdoc__
from .feature import FeatureDescriptor, MultipleFeatureDescriptors
from .feature_collection import FeatureCollection
from .function_wrapper import NumpyFuncWrapper
from .logger import get_feature_logs, get_function_stats, get_series_names_stats

__pdoc__["NumpyFuncWrapper.__call__"] = True

__all__ = [
    "FeatureDescriptor",
    "MultipleFeatureDescriptors",
    "FeatureCollection",
    "NumpyFuncWrapper",
    "get_feature_logs",
    "get_function_stats",
    "get_series_names_stats",
]




from abc import ABCMeta, abstractmethod

import pandas as pd
from imblearn.base import SamplerMixin

class SegmenterMixin(SamplerMixin, metaclass=ABCMeta):

    def _check_X_y(self, X, y):
        assert len(X) == len(y)

    @abstractmethod
    def _sample(self, X, y=None) -> pd.DataFrame:
        pass

    def _fit_resample(self, X, y):
        return self._sample(X, y)

    def fit_resample(self, X, y) -> pd.DataFrame:
        self._check_X_y(X, y)
        return self._fit_resample(X, y)

    def segment(self, X, y = None) -> pd.DataFrame:
        return self._sample(X, y)        



"""FeatureCollection class for bookkeeping and calculation of time-series features.

See Also
--------
Example notebooks and model serialization documentation.

"""

from __future__ import annotations  # Make typing work for the enclosing class

__author__ = "Jonas Van Der Donckt, Emiel Deprost, Jeroen Van Der Donckt"

import dill
import pandas as pd
import numpy as np

from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union
from pathos.multiprocessing import ProcessPool
from tqdm.auto import tqdm

from .feature import FeatureDescriptor, MultipleFeatureDescriptors
from .logger import logger
from .strided_rolling import StridedRolling
from ..features.function_wrapper import NumpyFuncWrapper
from ..utils.data import to_list, to_series_list, flatten
from ..utils.time import timedelta_to_str
from ..utils.logging import delete_logging_handlers, add_logging_handler


class FeatureCollection:
    """Create a FeatureCollection.

    Parameters
    ----------
    feature_descriptors : Union[FeatureDescriptor, MultipleFeatureDescriptors, List[Union[FeatureDescriptor, MultipleFeatureDescriptors]]], optional
        Initial (list of) feature(s) to add to collection, by default None

    """

    def __init__(
        self,
        feature_descriptors: Optional[
            Union[
                FeatureDescriptor,
                MultipleFeatureDescriptors,
                List[Union[FeatureDescriptor, MultipleFeatureDescriptors]],
            ]
        ] = None,
    ):
        # The feature collection is a dict with keys of type:
        #   tuple(tuple(str), float OR pd.timedelta, float OR pd.timedelta)
        # The outer tuple's values correspond to (series_key(s), window, stride)
        self._feature_desc_dict: Dict[
            Tuple[Tuple[str], pd.Timedelta, pd.Timedelta], List[FeatureDescriptor]
        ] = {}

        if feature_descriptors:
            self.add(feature_descriptors)

    def get_required_series(self) -> List[str]:
        """Return all required series names for this feature collection.

        Return the list of series names that are required in order to calculate all the
        features (defined by the `FeatureDescriptor` objects) of this feature
        collection.

        Returns
        -------
        List[str]
            List of all the required series names.

        """
        return list(
            set(flatten([fr_key[0] for fr_key in self._feature_desc_dict.keys()]))
        )

    @staticmethod
    def _get_collection_key(
        feature: FeatureDescriptor,
    ) -> Tuple[tuple, pd.Timedelta, pd.Timedelta]:
        # Note: `window` & `stride` properties can either be a pd.Timedelta or an int
        return feature.series_name, feature.window, feature.stride

    def _add_feature(self, feature: FeatureDescriptor):
        """Add a `FeatureDescriptor` instance to the collection.

        Parameters
        ----------
        feature : FeatureDescriptor
            The feature that will be added to this feature collection.

        """
        series_win_stride_key = self._get_collection_key(feature)
        if series_win_stride_key in self._feature_desc_dict.keys():
            self._feature_desc_dict[series_win_stride_key].append(feature)
        else:
            self._feature_desc_dict[series_win_stride_key] = [feature]

    def add(
        self,
        features: Union[
            FeatureDescriptor, 
            MultipleFeatureDescriptors,
            FeatureCollection,
            List[
                Union[FeatureDescriptor, MultipleFeatureDescriptors, FeatureCollection]
            ],
        ],
    ):
        """Add feature(s) to the FeatureCollection.

        Parameters
        ----------
        features : Union[FeatureDescriptor, MultipleFeatureDescriptors, FeatureCollection, List[Union[FeatureDescriptor, MultipleFeatureDescriptors, FeatureCollection]]]
            Feature(s) (containers) whose contained features will be added.

        Raises
        ------
        TypeError
            Raised when an item within `features` is not an instance of
            [`MultipleFeatureDescriptors`, `FeatureDescriptors`, `FeatureCollection`].

        """
        # Convert to list if necessary
        features = to_list(features)

        for feature in features:
            if isinstance(feature, MultipleFeatureDescriptors):
                self.add(feature.feature_descriptions)
            elif isinstance(feature, FeatureDescriptor):
                self._add_feature(feature)
            elif isinstance(feature, FeatureCollection):
                # List needs to be flattened
                self.add(list(flatten(feature._feature_desc_dict.values())))
            else:
                raise TypeError(f"type: {type(feature)} is not supported - {feature}")

    @staticmethod
    def _executor(idx: int):
        # global get_stroll_func
        stroll, function = get_stroll_func(idx)
        return stroll.apply_func(function)

    def _stroll_feat_generator(
        self, series_dict: Dict[str, pd.Series], window_idx: str, approve_sparsity: bool
    ) -> List[Tuple[StridedRolling, NumpyFuncWrapper]]:
        # --- Future work ---
        # We could also make the StridedRolling creation multithreaded
        # Very low priority because the STROLL __init__ is rather efficient!
        keys_wins_strides = list(self._feature_desc_dict.keys())
        lengths = np.cumsum(
            [len(self._feature_desc_dict[k]) for k in keys_wins_strides]
        )

        def get_stroll_function(idx):
            key_idx = np.searchsorted(lengths, idx, "right")  # right bc idx starts at 0
            key, win, stride = keys_wins_strides[key_idx]
            stroll = StridedRolling(
                data=[series_dict[k] for k in key],
                window=win,
                stride=stride,
                window_idx=window_idx,
                approve_sparsity=approve_sparsity,
            )
            feature = self._feature_desc_dict[keys_wins_strides[key_idx]][
                idx - lengths[key_idx]
            ]
            return stroll, feature.function

        return get_stroll_function

    def _get_stroll_feat_length(self) -> int:
        return sum(
            len(self._feature_desc_dict[k]) for k in self._feature_desc_dict.keys()
        )

    def calculate(
        self,
        data: Union[pd.Series, pd.DataFrame, List[Union[pd.Series, pd.DataFrame]]],
        return_df: Optional[bool] = False,
        window_idx: Optional[str] = "end",
        approve_sparsity: Optional[bool] = False,
        show_progress: Optional[bool] = False,
        logging_file_path: Optional[Union[str, Path]] = None,
        n_jobs: Optional[int] = None,
    ) -> Union[List[pd.DataFrame], pd.DataFrame]:
        """Calculate features on the passed data.

        Notes
        ------
        * The (column-)names of the series in `data` represent the names in the keys.
        * If a `logging_file_path` is provided, the execution (time) info can be
          retrieved by calling `logger.get_feature_logs(logging_file_path)`.
          Be aware that the `logging_file_path` gets cleared before the logger pushes
          logged messages. Hence, one should use a separate logging file for each
          constructed processing and feature instance with this library.

        Parameters
        ----------
        data : Union[pd.Series, pd.DataFrame, List[Union[pd.Series, pd.DataFrame]]]
            Dataframe or Series or list thereof, with all the required data for the
            feature calculation. \n
            **Remark**: each Series / DataFrame must have a `pd.DatetimeIndex`.
            **Remark**: we assume that each name / column is unique.
        return_df : bool, optional
            Whether the output needs to be a dataframe list or a DataFrame, by default
            False.
            If `True` the output dataframes will be merged to a DataFrame with an outer
            merge.
        window_idx : str, optional
            The window's index position which will be used as index for the
            feature_window aggregation. Must be either of: ['begin', 'middle', 'end'],
            by default 'end'. All features in this collection will use the same
            window_idx.
        approve_sparsity: bool, optional
            Bool indicating whether the user acknowledges that there may be sparsity
            (i.e., irregularly sampled data), by default False.
            If False and sparsity is observed, a warning is raised.
        show_progress: bool, optional
            If True, the progress will be shown with a progressbar, by default False.
        logging_file_path : Union[str, Path], optional
            The file path where the logged messages are stored. If `None`, then no
            logging `FileHandler` will be used and the logging messages are only pushed
            to stdout. Otherwise, a logging `FileHandler` will write the logged messages
            to the given file path.
        n_jobs : int, optional
            The number of processes used for the feature calculation. If `None`, then
            the number returned by `os.cpu_count()` is used, by default None. \n
            If n_jobs is either 0 or 1, the code will be executed sequentially without
            creating a process pool. This is very useful when debugging, as the stack
            trace will be more comprehensible.

            .. tip::
                * It takes on avg. _300ms_ to schedule everything with
                  multiprocessing. So if your feature extraction code runs faster than
                  ~1.5s, it might not be worth it to parallelize the process
                  (and thus better set the `n_jobs` to 0-1).
                * This method its memory peaks are significantly lower when executed
                  sequentially. Set the `n_jobs` to 0-1 if this matters.

        Returns
        -------
        Union[List[pd.DataFrame], pd.DataFrame]
            The calculated features.

        Raises
        ------
        KeyError
            Raised when a required key is not found in `data`.

        """
        # Delete other logging handlers
        delete_logging_handlers(logger)
        # Add logging handler (if path provided)
        if logging_file_path:
            add_logging_handler(logger, logging_file_path)

        # Convert the data to a series_dict
        series_dict: Dict[str, pd.Series] = {}
        for s in to_series_list(data):
            # Assert the assumptions we make!
            assert isinstance(s.index, pd.DatetimeIndex)
            assert s.index.is_monotonic_increasing

            if s.name in self.get_required_series():
                series_dict[str(s.name)] = s

        calculated_feature_list: List[pd.DataFrame] = []
        
        # Note: this variable has a global scope so this is shared in multiprocessing
        global get_stroll_func
        get_stroll_func = self._stroll_feat_generator(
            series_dict, window_idx, approve_sparsity
        )

        if n_jobs in [0, 1]:
            idxs = range(self._get_stroll_feat_length())
            if show_progress:
                idxs = tqdm(idxs)
            calculated_feature_list = [self._executor(idx) for idx in idxs]
        else:
            # https://pathos.readthedocs.io/en/latest/pathos.html#usage
            with ProcessPool(nodes=n_jobs, source=True) as pool:
                results = pool.uimap(
                    self._executor,
                    range(self._get_stroll_feat_length()),
                )
                if show_progress:
                    results = tqdm(results, total=self._get_stroll_feat_length())
                calculated_feature_list = [f for f in results]
                # Close & join - see: https://github.com/uqfoundation/pathos/issues/131
                pool.close()
                pool.join()
                # Clear because: https://github.com/uqfoundation/pathos/issues/111
                pool.clear()

        if return_df:
            return pd.concat(calculated_feature_list, axis=1, join="outer", copy=False)
        else:
            return calculated_feature_list

    def serialize(self, file_path: Union[str, Path]):
        """Serialize this `FeatureCollection` instance.

        Parameters
        ----------
        file_path : Union[str, Path]
            The path where the `FeatureCollection` will be serialized.

        Notes
        -----
        * As we use [Dill](https://github.com/uqfoundation/dill){:target="_blank"} to
          serialize the files, we can **also serialize functions which are defined in
          the local scope, like lambdas.**

        """
        with open(file_path, "wb") as f:
            dill.dump(self, f, recurse=True)

    def __repr__(self) -> str:
        """Representation string of a FeatureCollection."""
        feature_keys = sorted(set(k[0] for k in self._feature_desc_dict.keys()))
        output_str = ""
        for feature_key in feature_keys:
            output_str += f"{'|'.join(feature_key)}: ("
            keys = (x for x in self._feature_desc_dict.keys() if x[0] == feature_key)
            for _, win_size, stride in keys:
                output_str += f"\n\twin: "
                win_str = timedelta_to_str(win_size)
                stride_str = timedelta_to_str(stride)
                output_str += f"{str(win_str):<6}, stride: {str(stride_str)}: ["
                for feat_desc in self._feature_desc_dict[feature_key, win_size, stride]:
                    output_str += f"\n\t\t{feat_desc._func_str},"
                output_str += "\n\t]"
            output_str += "\n)\n"
        return output_str



# -*- coding: utf-8 -*-
"""(Advanced) tsflex utilities for chunking time-series data."""

__author__ = 'Jonas Van Der Donckt'

from datetime import timedelta
from typing import Dict, List, Union, Tuple, Optional

import pandas as pd

from ..utils.data import to_series_list
from ..utils.time import parse_time_arg


def chunk_data(
        data: Union[pd.Series, pd.DataFrame, List[Union[pd.Series, pd.DataFrame]]],
        fs_dict: Dict[str, int],
        chunk_range_margin: Optional[Union[float, str, pd.Timedelta]] = None,
        min_chunk_dur: Optional[Union[float, str, pd.Timedelta]] = None,
        max_chunk_dur: Optional[Union[float, str, pd.Timedelta]] = None,
        sub_chunk_overlap: Optional[Union[float, str, pd.Timedelta]] = "0s",
        copy=True,
        verbose=False,
) -> List[List[pd.Series]]:
    r"""Divide the time-series `data` in same time-range chunks.

    Does 2 things:

    1. Detecting gaps in the `data`(-list) time series.
    2. Divides the `data` into chunks, according to the parameter
        configuration and the detected gaps.

    Notes
    -----
    * As the `fs_dict` is requested, the assumption is made that **each item** in `data`
      has a **fixed sample frequency** and that **each item** in `data` 
    * All subsequent series-chunks are matched against the time-ranges of the first
      series. This implies that **the first item in `data` serves as a reference**
      for gap-matching.
    * The term `sub-chunk` refers to the chunks who exceed the `max_chunk_duration_s`
      parameter and are therefore further divided into sub-chunks.
      TODO: @Jeroen -> zo een note toevoegen aan constructor featuredescriptor

    Parameters
    ----------
    data: Union[pd.Series, pd.DataFrame, List[Union[pd.Series, pd.DataFrame]]]
        The time-series data which will be chunked. Each item in `data` must have a 
        `pd.DatetimeIndex`. 
        The assumption is made that each `item` in data has a _nearly-constant_ sample 
        frequency (when there are no gaps).
    fs_dict: Dict[str, int]
        The sample frequency dict. This dict must at least withhold all the keys
        from the items in `data`.
    chunk_range_margin: Union[float, str, pd.Timedelta], optional
        The allowed margin (in seconds if a float) between same time-range chunks their
        start and end time. If `None` the margin will be set as:

            2 / min(fs_dict.intersection(data.names).values())

         Which is equivalent to twice the min-fs (= max-period) of the passed `data`,
         by default None.
    min_chunk_dur : Union[float, str, pd.Timedelta], optional
        The minimal duration of a chunk (in seconds if a float), by default None
        Chunks with durations smaller than this will be discarded (and not returned).
    max_chunk_dur : Union[float, str, pd.Timedelta], optional
        The maximal duration of a chunk (in seconds if a float), by default None
        Chunks with durations larger than this will be chunked in smaller `sub_chunks`
        where each sub-chunk has a maximal duration of `max_chunk_dur_s`.
    sub_chunk_overlap: Union[float, str, pd.Timedelta], optional
        The sub-chunk boundary overlap (in seconds if a float). If available, this
        margin will be added to either side of the `sub_chunk`. \n
        This is especially useful to not lose inter-`sub_chunk` data (as each
        `sub_chunk` is in fact a continuous chunk) when window-based aggregations
        are performed on these same time range output (sub_)chunks. \n
        This argument is only relevant if `max_chunk_dur_s` is set.
    copy: boolean, optional
        If set True will return a new view (on which you won't get a
        `SettingWithCopyWarning` if you change the content), by default False.
    verbose : bool, optional
        If set, will print more verbose output, by default False

    Returns
    -------
    List[List[pd.Series]]
        A list of same time range chunks.

    """
    # Convert the input data
    series_list = to_series_list(data)

    if min_chunk_dur is not None:
        min_chunk_dur = parse_time_arg(min_chunk_dur)
    if max_chunk_dur is not None:
        max_chunk_dur = parse_time_arg(max_chunk_dur)
    sub_chunk_overlap = parse_time_arg(sub_chunk_overlap)

    # Assert that there are no duplicate series names and the names reside in fs_dict
    assert len(series_list) == len(set([s.name for s in series_list]))
    assert all([str(s.name) in fs_dict for s in series_list])

    # Default arg -> set the chunk range margin to 2x the min-freq its period
    if chunk_range_margin is None:
        chunk_range_margin = 2 / min([fs_dict[str(s.name)] for s in series_list])

    chunk_range_margin = parse_time_arg(chunk_range_margin)
    assert chunk_range_margin.total_seconds() > 0, "chunk_range_margin must be > 0"

    # Some range asserts
    assert sub_chunk_overlap.total_seconds() >= 0, f"sub_chunk_overlap_s must be > 0"
    if max_chunk_dur is not None:
        assert max_chunk_dur.total_seconds() > 0, f"max_chunk_dur_s must be > 0"

    # Variable in which the same time-range chunks are stored
    # Each list item can be seen as (t_start_chunk, t_end_chunk, chunk_list)
    same_range_chunks: List[Tuple[pd.Timestamp, pd.Timestamp, List[pd.Series]]] = []

    def print_verbose_time(sig, t_begin, t_end, msg=""):
        fmt = "%Y-%m-%d %H:%M"
        if not verbose:
            return
        print(
            f"slice {sig.name} {t_begin.strftime(fmt):<10} - {t_end.strftime(fmt):<10} "
            f"-  shape: {sig[t_begin:t_end].shape}"
        )
        if len(msg):
            print(f"\t└──>  {msg}")

    def slice_time(sig: pd.Series, t_begin: pd.Timestamp,
                   t_end: pd.Timestamp) -> pd.Series:
        """Slice the sig Series."""
        if copy:
            return sig[t_begin:t_end].copy()
        else:
            return sig[t_begin:t_end]

    def insert_chunk(chunk: pd.Series):
        """Insert the chunk into `same_range_chunks`."""
        t_chunk_start, t_chunk_end = chunk.index[[0, -1]]

        # Iterate over the same-(time)range-chunk (src) collection
        for src_start, src_end, src_chunks in same_range_chunks:
            # Check for overlap
            if (abs(src_start - t_chunk_start) <= chunk_range_margin and
                    abs(src_end - t_chunk_end) <= chunk_range_margin):
                # Check series name not in src_chunks
                if chunk.name not in [src_c.name for src_c in src_chunks]:
                    src_chunks.append(chunk)
                    print_verbose_time(chunk, t_chunk_start, t_chunk_end,
                                       "INSERT chunk")
                    return
                else:
                    # There already exists a sub_chunk of this series name
                    raise ValueError("There already exists a chunk with this series"
                                     f" name - {chunk.name}")

        same_range_chunks.append((t_chunk_start, t_chunk_end, [chunk]))
        print_verbose_time(chunk, t_chunk_start, t_chunk_end, "APPEND sub chunk")

    for series in series_list:
        # 1. Some checks
        if len(series) < 2:
            if verbose:
                print(f"Too small series: {series.name} - shape: {series.shape} ")
            continue

        # Allowed offset (in seconds) is sample_period + 0.5*sample_period
        fs_sig = fs_dict[str(series.name)]
        gaps = series.index.to_series().diff() > timedelta(seconds=(1 + .5) / fs_sig)
        # Set the first and last timestamp to True
        gaps.iloc[[0, -1]] = True
        gaps: List[pd.Timestamp] = series[gaps].index.to_list()
        if verbose:
            print("-" * 10, " detected gaps", "-" * 10)
            print(*gaps, sep="\n")

        for (t_begin_c, t_end_c) in zip(gaps, gaps[1:]):
            # The t_end is the t_start of the new time range -> hence [:-1]
            # => cut on [t_start_c(hunk), t_end_c(hunk)[
            sig_chunk = series[t_begin_c:t_end_c]
            if t_end_c < gaps[-1]:
                # Note: we doe not adjust when t_end_c = gaps[-1]
                #   (as gaps-[-1] is not really a gap)
                sig_chunk = sig_chunk[:-1]

            if len(sig_chunk) > 2:  # re-adjust the t_end
                t_end_c = sig_chunk.index[-1]
            else:
                print_verbose_time(series, t_begin_c, t_end_c, "too small df_chunk")
                continue

            # Check for min duration
            chunk_range_s = int(len(sig_chunk) / fs_sig)
            if min_chunk_dur is not None and chunk_range_s < \
                    min_chunk_dur.total_seconds():
                print_verbose_time(
                    sig_chunk,
                    t_begin_c,
                    t_end_c,
                    f"Too small chunk min_dur {min_chunk_dur} > {sig_chunk.shape}",
                )
                continue

            # Divide the chunk into sub_chunks (sc's)
            if max_chunk_dur is not None and chunk_range_s > \
                    max_chunk_dur.total_seconds():
                print_verbose_time(
                    sig_chunk, t_begin_c, t_end_c, "Dividing in sub-chunks"
                )
                t_begin_sc = t_begin_c
                while t_begin_sc < t_end_c:
                    # Calculate the end sub-chunk time
                    t_end_sc = t_begin_sc + max_chunk_dur

                    # Get the end and begin sub-chunk margin
                    t_end_sc_m = t_end_sc + sub_chunk_overlap
                    t_end_sc_m = min(t_end_c, t_end_sc_m)

                    t_begin_sc_m = t_begin_sc - sub_chunk_overlap
                    t_begin_sc_m = max(t_begin_c, t_begin_sc_m)

                    # Slice & add the sub-chunk to the list
                    insert_chunk(chunk=slice_time(series, t_begin_sc_m, t_end_sc_m), )

                    # Update the condition's variable
                    t_begin_sc = t_end_sc
            else:
                insert_chunk(chunk=slice_time(series, t_begin_c, t_end_c))
    return [chunk_range[-1] for chunk_range in same_range_chunks]



"""Contains a (rather) fast implementation of a **time-based** strided rolling window.

.. todo::
    look into **series-based** stroll, instead of np.ndarray based stroll.<br>
    advantages:\n
    * a series is a wrapper around a 1D np.ndarray, so all np-based operations should
      work
    * the end-user can always use the time-index for advanced feature calculation e.g.
      window-based delayed correlation or something like that.

"""

__author__ = "Jonas Van Der Donckt, Jeroen Van Der Donckt, Emiel Deprost"

import time
import warnings
import pandas as pd
import numpy as np

from collections import namedtuple
from typing import Callable, Union, List, Tuple, Optional

from .function_wrapper import NumpyFuncWrapper
from .logger import logger
from ..utils.data import to_series_list
from ..utils.time import timedelta_to_str


class StridedRolling:
    """Custom time-based sliding window with stride for pandas DataFrames.

    Parameters
    ----------
    data : Union[pd.Series, pd.DataFrame]
        ``pd.Series`` or ``pd.DataFrame`` to slide over, the index must be a
        (time-zone-aware) ``pd.DatetimeIndex``.
    window : Union[int, pd.Timedelta]
        Either an int or ``pd.Timedelta``, representing the sliding window length in
        samples or the sliding window duration, respectively.
    stride : Union[int, pd.Timedelta]
        Either an int or ``pd.Timedelta``, representing the stride size in samples or
        the stride duration, respectively.
    window_idx : str, optional
        The window's index position which will be used as index for the
        feature_window aggregation. Must be either of: ['begin', 'middle', 'end'], by
        default 'end'.
    bound_method: str, optional
        The start-end bound methodology which is used to generate the slice ranges when
        ``data`` consists of multiple series / columns.
        Must be either of: ['inner', 'outer', 'first'], by default 'inner'.

        * if ``inner``, the inner-bounds of the series are used, the
        * if ``outer``, the inner-bounds of the series are used
        * if ``first``, the first-series it's bound will be used
    approve_sparsity: bool, optional
        Bool indicating whether the user acknowledges that there may be sparsity (i.e.,
        irregularly sampled data), by default False.
        If False and sparsity is observed, a warning is raised.

    Notes
    -----
    * This instance withholds a **read-only**-view of the data its values.

    <br>

    .. todo::
        The `bound_method`-argument must still be propagated to `FeatureCollection`

    """

    # Create the named tuple
    _NumpySeriesContainer = namedtuple(
        "SeriesContainer", ["values", "start_indexes", "end_indexes"]
    )

    def __init__(
            self,
            data: Union[pd.Series, pd.DataFrame, List[Union[pd.Series, pd.DataFrame]]],
            window: pd.Timedelta,
            stride: pd.Timedelta,
            window_idx: Optional[str] = "end",
            bound_method: Optional[str] = "inner",
            approve_sparsity: Optional[bool] = False,
    ):
        self.window: pd.Timedelta = window
        self.stride: pd.Timedelta = stride

        # 0. standardize the input
        series_list: List[pd.Series] = to_series_list(data)
        self.series_key: Tuple[str] = tuple([str(s.name) for s in series_list])

        # 1. Determine the bounds
        t_start, t_end = self._determine_bounds(series_list, bound_method)

        # And slice **all** the series to these tightest bounds
        assert (t_end - t_start) > window
        if len(series_list) > 1:
            series_list = [s[t_start:t_end] for s in series_list]

        # 2. Create the time_index which will be used for DataFrame reconstruction
        # 2.1 - and adjust the time_index
        # note: this code can also be placed in the `apply_func` method (if we want to
        #  make the bound window-idx setting feature specific).
        if window_idx == "end":
            window_idx_offset = window
        elif window_idx == "middle":
            window_idx_offset = window / 2
        elif window_idx == "begin":
            window_idx_offset = pd.Timedelta(seconds=0)
        else:
            raise ValueError(
                f"window index {window_idx} must be either of: "
                "['end', 'middle', 'begin']"
            )

        # use closed = left to exclude 'end' if it falls on the boundary
        # note: the index automatically takes the timezone of `t_start` & `t_end`
        # note: the index-name of the first passed series will be used
        self.index = pd.date_range(
            start=t_start + window_idx_offset,
            end=t_end - window + window_idx_offset,
            freq=stride,
            name=series_list[0].index.name
        )

        # ---------- Efficient numpy code -------
        # 1. Convert everything to int64
        np_start = t_start.to_datetime64()
        np_window = self.window.to_timedelta64()
        np_stride = self.stride.to_timedelta64()

        # 2. Precompute the start & end times (these remain the same for each series)
        # note: this if equivalent to:
        #   if `window` == 'begin":
        #       start_times = self.index.values
        np_start_times = np.arange(
            start=np_start, stop=np_start + (len(self.index) * np_stride),
            step=np_stride,
            dtype=np.datetime64,
        )
        np_end_times = np_start_times + np_window

        self.series_containers: List[StridedRolling._NumpySeriesContainer] = []
        for series in series_list:
            # create a non-writeable view of the series
            np_series = series.values
            np_series.flags.writeable = False

            np_idx_times = series.index.values
            self.series_containers.append(
                StridedRolling._NumpySeriesContainer(
                    # TODO: maybe save the pd.Series instead of the np.series
                    values=np_series,
                    # the slicing will be performed on [ t_start, t_end [
                    # TODO: this can maybe be optimized -> further look into this
                    # np_idx_times, np_start_times, & np_end_times are all sorted!
                    # as we assume & check that the time index is monotonically
                    # increasing & the latter 2 are created using `np.arange()`
                    start_indexes=np.searchsorted(np_idx_times, np_start_times, "left"),
                    end_indexes=np.searchsorted(np_idx_times, np_end_times, "left"),
                )
            )

            if not approve_sparsity:
                last_container = self.series_containers[-1]
                qs = [0, 0.1, 0.5, 0.9, 1]
                series_idx_stats = np.quantile(
                    last_container.end_indexes - last_container.start_indexes, q=qs
                )
                q_str = ", ".join([f"q={q}: {v}" for q, v in zip(qs, series_idx_stats)])
                if not all(series_idx_stats == series_idx_stats[-1]):  # min != max
                    warnings.warn(
                        f"There are gaps in the time-series {series.name}; "
                        + f"\n \t Quantiles of nb values in window: {q_str}",
                        RuntimeWarning,
                    )

    @staticmethod
    def _determine_bounds(
            series_list: List[pd.Series], bound_method: str
    ) -> Tuple[pd.Timestamp, pd.Timestamp]:
        """Determine the bounds of the passed series.

        Parameters
        ----------
        series_list : List[pd.Series]
            The list of series for which the bounds are determined.

        bound_method : str
            The methodology which is used for the ``series_list`` bound determination

        Returns
        -------
        Tuple[pd.Timestamp, pd.Timestamp]
            The start & end timestamp, respectively.

        """
        if bound_method == "inner":
            latest_start = series_list[0].index[0]
            earliest_stop = series_list[0].index[-1]
            for series in series_list[1:]:
                latest_start = max(latest_start, series.index[0])
                earliest_stop = min(earliest_stop, series.index[-1])
            return latest_start, earliest_stop

        if bound_method == "outer":
            earliest_start = series_list[0].index[0]
            latest_stop = series_list[0].index[-1]
            for series in series_list[1:]:
                earliest_start = min(earliest_start, series.index[0])
                latest_stop = max(latest_stop, series.index[-1])
            return earliest_start, latest_stop

        elif bound_method == "first":
            return series_list[0].index[0], series_list[0].index[-1]

        else:
            raise ValueError(f"invalid bound method string passed {bound_method}")

    def apply_func(self, np_func: Union[Callable, NumpyFuncWrapper]) -> pd.DataFrame:
        """Apply a function to the expanded time-series.

        Parameters
        ----------
        np_func : Union[Callable, NumpyFuncWrapper]
            The Callable (wrapped) function which will be applied.

        Returns
        -------
        pd.DataFrame
            The merged output of the function applied to every column in a
            new DataFrame. The DataFrame's column-names have the format:
                `<series_col_name(s)>_<feature_name>__w=<window>_s=<stride>`.

        Raises
        ------
        ValueError
            If the passed ``np_func`` tries to adjust the data its read-only view.

        Notes
        -----
        * If ``np_func`` is only a callable argument, with no additional logic, this
          will only work for a one-to-one mapping, i.e., no multiple feature-output
          columns are supported for this case!<br>
          If you want to calculate one-to-many, ``np_func`` should be
          a ``NumpyFuncWrapper`` instance and explicitly use
          the ``output_names`` attributes of its constructor.

        """

        # Convert win & stride to time-string if available :)
        def create_feat_col_name(feat_name) -> str:
            win_str = timedelta_to_str(self.window)
            stride_str = timedelta_to_str(self.stride)
            win_stride_str = f"w={win_str}_s={stride_str}"
            return f"{'|'.join(self.series_key)}__{feat_name}__{win_stride_str}"

        if not isinstance(np_func, NumpyFuncWrapper):
            np_func = NumpyFuncWrapper(np_func)
        feat_names = np_func.output_names

        t_start = time.time()

        # --- Future work ---
        # would be nice if we could optimize this double for loop with something
        # more vectorized
        out = np.array(
            [np_func(
                *[sc.values[sc.start_indexes[idx]: sc.end_indexes[idx]]
                  for sc in self.series_containers]
            ) for idx in range(len(self.index))]
        )

        # Aggregate function output in a dictionary
        feat_out = {}
        if out.ndim == 1 or (out.ndim == 2 and out.shape[1] == 1):
            assert len(feat_names) == 1
            feat_out[create_feat_col_name(feat_names[0])] = out.flatten()
        if out.ndim == 2 and out.shape[1] > 1:
            assert len(feat_names) == out.shape[1]
            for col_idx in range(out.shape[1]):
                feat_out[create_feat_col_name(feat_names[col_idx])] = out[:, col_idx]

        elapsed = time.time() - t_start
        logger.info(
            f"Finished function [{np_func.func.__name__}] on "
            f"{[self.series_key]} with window-stride "
            f"[{self.window}, {self.stride}] in [{elapsed} seconds]!"
        )

        return pd.DataFrame(index=self.index, data=feat_out)



# -*- coding: utf-8 -*- # TODO: rename file
"""(Advanced) utilities for the processing pipelines."""

__author__ = "Jonas Van Der Donckt, Jeroen Van Der Donckt"

import os
import traceback
from typing import List, Any, Optional, Union

import pandas as pd
from pathos.multiprocessing import ProcessPool
from tqdm.auto import tqdm

from .series_pipeline import SeriesPipeline


def process_chunks_multithreaded(
        same_range_chunks_list: List[List[Union[pd.Series, pd.DataFrame]]],
        series_pipeline: SeriesPipeline,
        show_progress: Optional[bool] = True,
        n_jobs: Optional[int] = None,
        **processing_kwargs,
) -> List[Any]:
    """Process `same_range_chunks_list` in a multithreaded manner, order is preserved.

    Parameters
    ----------
    same_range_chunks_list: List[List[Union[pd.Series, pd.DataFrame]]]
        A list of same-range-chunks, most likely the output of `chunk_data`.
    series_pipeline: SeriesPipeline
        The pipeline that will process each item in `same_range_chunks_list`.
    show_progress: bool, optional
        If True, the progress will be shown with a progressbar, by default True.
    n_jobs: int, optional
        The number of processes used for the chunked series processing. If `None`, then
        the number returned by `os.cpu_count()` is used, by default None.
    **processing_kwargs
        Keyword arguments that will be passed on to the processing pipeline.

    Returns
    -------
    List[Any]
        A list of the `series_pipeline` its processed outputs. The order is preserved.

    Notes
    -----
    * This method is not concerned with joining the chunks as this operation is highly
      dependent on the preprocessing steps. This is the user's responsibility.
    * If any error occurs while executing the `series_pipeline` on one of the chunks
      in `same_range_chunks_list`, the traceback is printed and an empty dataframe is
      returned. We chose for this behavior, because in this way the other parallel
      processes are not halted in case of an error.

    """
    if n_jobs is None:
        n_jobs = os.cpu_count()

    def _executor(same_range_chunks: List[Union[pd.Series, pd.DataFrame]]):
        try:
            return series_pipeline.process(same_range_chunks, **processing_kwargs)
        except Exception:
            # Print traceback and return empty `pd.DataFrame` in order to not break the
            # other parallel processes.
            traceback.print_exc()
            return pd.DataFrame()

    processed_out = []
    with ProcessPool(nodes=min(n_jobs, len(same_range_chunks_list)),
                     source=True) as pool:
        results = pool.imap(_executor, same_range_chunks_list)
        if show_progress:
            results = tqdm(results, total=len(same_range_chunks_list))
        for f in results:
            processed_out.append(f)
        # Close & join because: https://github.com/uqfoundation/pathos/issues/131
        pool.close()
        pool.join()
        # Clear because: https://github.com/uqfoundation/pathos/issues/111
        pool.clear()
    return processed_out



"""Contains the used variables and functions to provide logging functionality.

See Also
--------
FeatureCollection: its `logging_file_path` of the `calculate` method.

"""

__author__ = "Jeroen Van Der Donckt"

import logging
import pandas as pd
import re

from ..utils.logging import logging_file_to_df
from ..utils.time import timedelta_to_str

# Package specific logger
logger = logging.getLogger("feature_calculation_logger")
logger.setLevel(logging.DEBUG)

# Create logger which writes WARNING messages or higher to sys.stderr
console = logging.StreamHandler()
console.setLevel(logging.WARNING)
logger.addHandler(console)


def _parse_message(message: str) -> list:
    """Parse the message of the logged info."""
    regex = r"\[(.*?)\]"
    matches = re.findall(regex, message)
    assert len(matches) == 4
    func = matches[0]
    key = matches[1].replace("'", "") 
    window, stride = matches[2].split(",")[0], matches[2].split(",")[1]
    duration_s = float(matches[3].rstrip(" seconds"))
    return [func, key, window, stride, duration_s]


def _parse_logging_execution_to_df(logging_file_path: str) -> pd.DataFrame:
    """Parse the logged messages into a dataframe that contains execution info.

    Parameters
    ----------
    logging_file_path: str
        The file path where the logged messages are stored. This is the file path that
        is passed to the `FeatureCollection` its `calculate` method. 

    Note
    ----
    This function only works when the `logging_file_path` used in a
    `FeatureCollection` its `calculate` method is passed.

    Returns
    -------
    pd.DataFrame
        A DataFrame with the features its function, input series names and 
        calculation duration.

    """
    df = logging_file_to_df(logging_file_path)
    df[["function", "series_names", "window", "stride", "duration"]] = list(
        df["message"].apply(_parse_message)
    )
    df["window"] = pd.to_timedelta(df["window"]).apply(timedelta_to_str)
    df["stride"] = pd.to_timedelta(df["stride"]).apply(timedelta_to_str)
    return df.drop(columns=["name", "log_level", "message"])


def get_feature_logs(logging_file_path: str) -> pd.DataFrame:
    """Get execution (time) info for each feature of a `FeatureCollection`.

    Parameters
    ----------
    logging_file_path: str
        The file path where the logged messages are stored. This is the file path that
        is passed to the `FeatureCollection` its `calculate` method.

    Returns
    -------
    pd.DataFrame
        A DataFrame with the features its function, input series names and 
        calculation duration.

    """
    df = _parse_logging_execution_to_df(logging_file_path)
    df["duration"] = pd.to_timedelta(df["duration"], unit="s")
    return df


def get_function_stats(logging_file_path: str) -> pd.DataFrame:
    """Get execution (time) statistics for each function.

    Parameters
    ----------
    logging_file_path: str
        The file path where the logged messages are stored. This is the file path that
        is passed to the `FeatureCollection` its `calculate` method.

    Returns
    -------
    pd.DataFrame
        A DataFrame with for each function (i.e., `function-(window,stride)`) 
        combination the mean (time), std (time), sum (time), and number of executions.

    """
    df = _parse_logging_execution_to_df(logging_file_path)
    return (
        df.groupby(["function", "window", "stride"])
        .agg({"duration": ["mean", "std", "sum", "count"]})
        .sort_values(by=("duration", "mean"), ascending=False)
    )


def get_series_names_stats(logging_file_path: str) -> pd.DataFrame:
    """Get execution (time) statistics for each `key-(window,stride)` combination.

    Parameters
    ----------
    logging_file_path: str
        The file path where the logged messages are stored. This is the file path that
        is passed to the `FeatureCollection` its `calculate` method.

    Returns
    -------
    pd.DataFrame
        A DataFrame with for each function the mean (time), std (time), sum (time), and
        number of executions.

    """
    df = _parse_logging_execution_to_df(logging_file_path)
    return (
        df.groupby(["series_names", "window", "stride"])
        .agg({"duration": ["sum", "mean", "std", "count"]})
        .sort_values(by=("duration", "sum"), ascending=False)
    )



"""SeriesPipeline class for time-series data (pre-)processing pipeline."""

__author__ = "Jonas Van Der Donckt, Emiel Deprost, Jeroen Van Der Donckt"

import dill
import pandas as pd

from pathlib import Path
from typing import Dict, List, Optional, Union

from .series_processor import SeriesProcessor
from .logger import logger
from ..utils.data import series_dict_to_df, to_series_list, flatten
from ..utils.logging import delete_logging_handlers, add_logging_handler


class _ProcessingError(Exception):
    pass


class SeriesPipeline:
    """Pipeline for applying ``SeriesProcessor`` objects sequentially.

    Parameters
    ----------
    processors : List[SeriesProcessor], optional
        List of ``SeriesProcessor`` objects that will be applied sequentially to the
        internal series dict, by default None. **The processing steps will be
        executed in the same order as passed in this list**.

    """

    def __init__(self, processors: Optional[List[SeriesProcessor]] = None):
        self.processing_steps: List[SeriesProcessor] = []  # TODO: dit private of niet?
        if processors is not None:
            self.processing_steps = processors

    def get_required_series(self) -> List[str]:
        """Return all required series names for this pipeline.

        Return the list of series names that are required in order to execute all the
        ``SeriesProcessor`` objects of this processing pipeline.

        Returns
        -------
        List[str]
            List of all the required series names.

        """
        return list(
            set(flatten([step.get_required_series() for step in self.processing_steps]))
        )

    def append(self, processor: SeriesProcessor) -> None:
        """Append a ``SeriesProcessor`` at the end of the pipeline.

        Parameters
        ----------
        processor : SeriesProcessor
            The ``SeriesProcessor`` that will be added to the end of the pipeline

        """
        self.processing_steps.append(processor)

    def insert(self, idx: int, processor: SeriesProcessor) -> None:
        """Insert a ``SeriesProcessor`` at the given index in the pipeline.

        Parameters
        ----------
        idx : int
            The index where the given processor should be inserted in the pipeline.
            Index 0 will insert the given processor at the front of the pipeline,
            and index ``len(pipeline)`` is equivalent to appending the processor.
        processor : SeriesProcessor
            The ``SeriesProcessor`` that will be added to the end of the pipeline

        """
        self.processing_steps.insert(idx, processor)

    def process(
        self,
        data: Union[pd.Series, pd.DataFrame, List[Union[pd.Series, pd.DataFrame]]],
        return_df: Optional[bool] = False,
        return_all_series: Optional[bool] = True,
        drop_keys: Optional[List[str]] = None,
        copy: Optional[bool] = False,
        logging_file_path: Optional[Union[str, Path]] = None,
    ) -> Union[List[pd.Series], pd.DataFrame]:
        """Execute all ``SeriesProcessor`` objects in pipeline sequentially.

        Apply all the processing steps on passed Series list or DataFrame and return the
        preprocessed Series list or DataFrame.

        Parameters
        ----------
        data : Union[pd.Series, pd.DataFrame, List[Union[pd.Series, pd.DataFrame]]]
            Dataframe or Series or list thereof, with all the required data for the
            processing steps. \n
            **Remark**: each Series / DataFrame must have a ``pd.DatetimeIndex``.
            **Remark**: we assume that each name / column is unique.
        return_df : bool, optional
            Whether the output needs to be a series list or a DataFrame, by default 
            False.
            If True the output series will be combined to a DataFrame with an outer
            merge.
        return_all_series : bool, optional
            Whether the output needs to return all the series, by default True.
            * If True the output will contain all series that were passed to this
            method.
            * If False the output will contain just the required series (see
            ``get_required_series``).
        drop_keys : List[str], optional
            Which keys should be dropped when returning the output, by default None.
        copy : bool, optional
            Whether the series in ``data`` should be copied, by default False.
        logging_file_path : Union[str, Path], optional
            The file path where the logged messages are stored, by default None.
            If ``None``, then no logging ``FileHandler`` will be used and the logging
            messages are only pushed to stdout. Otherwise, a logging ``FileHandler`` will
            write the logged messages to the given file path.

        Returns
        -------
        Union[List[pd.Series], pd.DataFrame]
            The preprocessed series.

        Notes
        -----
        * If a ``logging_file_path`` is provided, the execution (time) info can be
          retrieved by calling ``logger.get_processor_logs(logging_file_path)``. <br>
          Be aware that the ``logging_file_path`` gets cleared before the logger pushes
          logged messages. Hence, one should use a separate logging file for each
          constructed processing and feature instance with this library.
        * If a series processor its function output is a ``np.ndarray``, the input series
          dict (required dict for that function) must contain just 1 series! That series
          its name and index are used to return a series dict. When a user does not want
          a numpy array to replace its input series, it is his / her responsibility to
          create a new ``pd.Series`` (or ``pd.DataFrame``) of that numpy array with a
          different (column) name.
        * If ``func_output`` is a ``pd.Series``, keep in mind that the input series gets
          transformed (i.e., replaced) in the pipeline with the ``func_output`` when the
          series name is  equal.

        Raises
        ------
        _ProcessingError
            Error raised when a processing step fails.

        """
        # Delete other logging handlers
        delete_logging_handlers(logger)
        # Add logging handler (if path provided)
        if logging_file_path:
            add_logging_handler(logger, logging_file_path)

        # Convert the data to a series_dict
        series_dict: Dict[str, pd.Series] = {}
        for s in to_series_list(data):
            # Assert the assumptions we make!
            if len(s):
                assert isinstance(s.index, pd.DatetimeIndex)
            # TODO: also check monotonic increasing?

            if s.name in self.get_required_series():
                series_dict[str(s.name)] = s.copy() if copy else s
            elif return_all_series:
                # If all the series have to be returned
                series_dict[str(s.name)] = s.copy() if copy else s

        output_keys = set()  # Maintain set of output series
        for processor in self.processing_steps:
            try:
                processed_dict = processor(series_dict)
                output_keys.update(processed_dict.keys())
                series_dict.update(processed_dict)
            except Exception as e:
                raise _ProcessingError(
                    "Error while processing function {}:\n {}".format(
                        processor.name, str(e)
                    )
                ) from e

        if not return_all_series:
            # Return just the output series
            output_dict = {key: series_dict[str(key)] for key in output_keys}
            series_dict = output_dict

        if drop_keys is not None:
            # Drop the keys that should not be included in the output
            output_dict = {
                key: series_dict[key]
                for key in set(series_dict.keys()).difference(drop_keys)
            }
            series_dict = output_dict

        if return_df:
            # We merge the series dict into a DataFrame
            return series_dict_to_df(series_dict)
        else:
            return [s for s in series_dict.values()]

    def serialize(self, file_path: Union[str, Path]):
        """Serialize this ``SeriesPipeline`` instance.

        Notes
        ------
        As we use [Dill](https://github.com/uqfoundation/dill){:target="_blank"} to
        serialize, we can also serialize (decorator)functions which are defined in the
        local scope, like lambdas.

        Parameters
        ----------
        file_path : Union[str, Path]
            The path where the ``SeriesProcessor`` will be serialized.

        """
        with open(file_path, "wb") as f:
            dill.dump(self, f, recurse=True)

    def __repr__(self):
        """Return formal representation of object."""
        return (
            "[\n" + "".join([f"\t{str(p)}\n" for p in self.processing_steps]) + "]"
        )

    def __str__(self):
        """Return informal representation of object."""
        return self.__repr__()



"""Code for time-series data (pre-)processing."""

__author__ = "Jonas Van Der Donckt, Emiel Deprost, Jeroen Van Der Donckt"

import time
from typing import Callable, Dict, List, Union, Tuple

import numpy as np
import pandas as pd

from .. import __pdoc__
from ..utils.classes import FrozenClass
from .logger import logger
from ..utils.data import series_dict_to_df, to_list, to_tuple, flatten

__pdoc__['SeriesProcessor.__call__'] = True


def dataframe_func(func: Callable):
    """Decorate function to use a DataFrame instead of multiple series (as argument).

    This decorator can be used for functions that need to work on a whole
    `pd.DataFrame`. It will convert the required series into a DataFrame using an
    **outer merge**.

    The function's prototype should be:

        func(df : pd.DataFrame, **kwargs)
            -> Union[np.ndarray, pd.Series, pd.DataFrame, List[pd.Series]]

    So the decorated `func` has to take a DataFrame as first argument.

    Notes
    -----
    Only when you want to perform row-based operations, such as `df.dropna(axis=0)`,
    this wrapper is needed.
    Hence, in most cases that `func` requires a `pd.DataFrame`, series arguments would
    be sufficient; as you can perform column-based operations on multiple `pd.Series`
    (e.g., subtract 2 series) and most dataframe operations are also available for a
    `pd.Series`.

    """

    def wrapper(*series: pd.Series, **kwargs):
        series_dict = {s.name: s for s in series}
        df = series_dict_to_df(series_dict)
        res = func(df, **kwargs)
        return res

    wrapper.__name__ = "dataframe_func: " + func.__name__
    return wrapper


class SeriesProcessor(FrozenClass):
    """Class that executes a specific operation on the passed series_dict.

    Parameters
    ----------
    function : Callable
        The function that processes the series (given in the `series_names`).
        The prototype of the function should match: \n

            function(*series: pd.Series, **kwargs)
                -> Union[np.ndarray, pd.Series, pd.DataFrame, List[pd.Series]]

        .. note::
            A function that processes a ``np.ndarray`` instead of ``pd.Series``
            should work just fine.

    series_names : Union[str, Tuple[str], List[str], List[Tuple[str]]]
        The names of the series on which the processing function should be applied.

        This argument should match the `function` its input; \n
        * If `series_names` is a (list of) string (or tuple of a single string),
          than `function` should require just one series as input.
        * If `series_names` is a (list of) tuple of strings, than `function` should
          require `len(tuple)` series as input **and in exactly the same order**.

        A list means multiple series (combinations) to process; \n
        * If `series_names` is a string or a tuple of strings, than `function` will
          be called only once for the series of this argument.
        * If `series_names` is a list of either strings or tuple of strings, than
          `function` will be called for each entry of this list.

        .. note::
            when passing a list as `series_names`, all items in this list should
            have the same type, i.e, either \n
            * all a `str`
            * or, all a `tuple` _with same length_. \n

    **kwargs: dict, optional
        Keyword arguments which will be also passed to the `function`

    Notes
    -----
    If the output of `function` is a `np.ndarray` or a `pd.Series` without a name,
    than (items of) the given `series_names` must have length 1, i.e., the function
    requires just 1 series! That series its name and index are used to transform
    (i.e., **replace**) that series.

    If you want to transform (i.e., **replace**) the input series with the
    processor, than `function` should return either: \n
    * a `np.ndarray` (see above).
    * a `pd.Series` with no name or with the same name as the input series.
    * a `pd.DataFrame` with (one) column name equal to the input series its name.
    * a list of `pd.Series` in which (exact) one series has the same name as the
      input series.

    Series (& columns) with other (column) names will be added to the series dict.

    """
    def __init__(
        self,
        function: Callable,
        series_names: Union[str, Tuple[str], List[str], List[Tuple[str]]],
        **kwargs,
    ):
        series_names = [to_tuple(names) for names in to_list(series_names)]
        # Assert that function inputs (series) all have the same length
        assert all(
            len(series_names[0]) == len(series_name_tuple)
            for series_name_tuple in series_names
        )
        self.series_names: List[Tuple[str]] = series_names
        self.function = function
        self.name = self.function.__name__

        self.kwargs = kwargs
        self._freeze()

    def get_required_series(self) -> List[str]:
        """Return all required series names for this processor.

        Return the list of series names that are required in order to execute the
        processing function.

        Returns
        -------
        List[str]
            List of all the required series names.

        """
        return list(set(flatten([name for name in self.series_names])))

    def __call__(self, series_dict: Dict[str, pd.Series]) -> Dict[str, pd.Series]:
        """**Call**culates the processed series.

        Parameters
        ----------
        series_dict : Dict[str, pd.Series]
            A dict of `pd.Series` containing the data that need to be processed.
            The key should always be the accompanying series its name.

        Returns
        -------
        Dict[str, pd.Series]
            The processed `series_dict`.

        Raises
        ------
        KeyError
            Raised when a key is not present in the `series_dict` but required for the
            processing.
        TypeError
            Raised when the output of the `SeriesProcessor` is not of the correct type.

        Notes
        -----
        * The `series_dict` is an internal representation of the time-series data .
          This internal representation is constructed in the `process` method of the
          `SeriesPipeline`.
        * If you want to test or debug your `SeriesProcessor` object, just encapsulate
          your instance of this class in a `SeriesPipeline`. The latter allows more
          versatile input for its `process` method.

        """
        t_start = time.time()

        # Only selecting the series that are needed for this processing step
        # requested_dict = {}
        # try:
        #     for sig in self.get_required_series():
        #         requested_dict[sig] = series_dict[sig]
        # except KeyError as key:
        #     # Re raise error as we can't continue
        #     raise KeyError(
        #         "Key %s is not present in the input dict and needed for processor %s"
        #         % (key, self.name)
        #     )

        # Variable that will contain the final output of this method
        processed_output: Dict[str, pd.Series] = {}

        def get_series_list(keys: Tuple[str]):
            """Get an ordered series list view for the given keys."""
            return [series_dict[key] for key in keys]

        def get_series_dict(keys: Tuple[str]):
            """Get a series dict view for the given keys."""
            return {key: series_dict[key] for key in keys}

        for series_name_tuple in self.series_names:
            func_output = self.function(
                *get_series_list(series_name_tuple), **self.kwargs
            )
            func_output = _handle_seriesprocessor_func_output(
                func_output,
                get_series_dict(series_name_tuple),
                self.name,
            )
            # Check that the output of the function call produces unique columns / keys
            assert (
                len(set(processed_output.keys()).intersection(func_output.keys())) == 0
            )
            processed_output.update(func_output)

        elapsed = time.time() - t_start
        logger.info(
            f"Finished function [{self.name}] on {self.series_names} in "
            f"[{elapsed} seconds]!"
        )

        return processed_output

    def __repr__(self):
        """Return formal representation of object."""
        repr_str = self.name + (" " + str(self.kwargs))
        repr_str += " :  " + " ".join([str(s) for s in self.series_names])
        return repr_str

    def __str__(self):
        """Return informal representation of object."""
        return self.__repr__()


# ---------------------- utility functions for a SeriesProcessor ----------------------

def _np_array_to_series(np_array: np.ndarray, series: pd.Series) -> pd.Series:
    """Convert the `np_array` into a pandas Series.

    Parameters
    ----------
    np_array: np.ndarray
        The numpy array that needs to be converted to a pandas Series.
    series: pd.Series
        The pandas Series that contains the name and the index for the `np_array`.

    Returns
    -------
    pd.Series
        A view of the numpy array as a pandas Series with index the given series
        its index and name the series its name.

    Notes
    -----
    * The constructed series is a pass-by-reference `view` of the numpy array.
    * The given `np_array` receives the same index and name as the `series`.
      Hence, the `np_array` needs to have the same length as the `series`.
    * Giving the `np_array` the **same name as** the `series`, will **result in**
      transforming (i.e., **replacing**) the `series` in the pipeline.
    * When a user does not want a numpy array to replace its input series, it is
      his / her responsibility to create a `pd.Series` (or `pd.DataFrame`) of that
      numpy array with a different (column) name.

    """
    # The length of the out has to be the same as the series length
    assert len(np_array) == len(series)
    # Note: pd.Series output holds a view of np_array!
    return pd.Series(data=np_array, index=series.index, name=series.name)


def _handle_seriesprocessor_func_output(
    func_output: Union[np.ndarray, pd.Series, pd.DataFrame, List[pd.Series]],
    required_dict: Dict[str, pd.Series],
    func_name: str,
) -> Union[Dict[str, pd.Series], pd.DataFrame]:
    """Handle the output of a SeriesProcessor its function.

    Parameters
    ----------
    func_output : Union[np.ndarray, pd.Series, pd.DataFrame, List[pd.Series]]
        The output of the SeriesProcessor its function.
    required_dict : Dict[str, pd.Series]
        The series dict that contains the required series for the `SeriesProcessor` its
        function.
    func_name : str
        The name of the SeriesProcessor (its function).

    Returns
    -------
    Union[Dict[str, pd.Series], pd.DataFrame]
        The processed outputs in the given func_output. \n
        * If the `func_output` is a `pd.DataFrame`, the `pd.DataFrame` is returned.
        * If the `func_output` is a `pd.Series`, `np.ndarray`, or list of `pd.Series`,
          a series dict is returned.

    Raises
    ------
    TypeError
        Error raised when function output is invalid.

    Notes
    -----
    * If `func_output` is a `np.ndarray` or a `pd.Series` without a name, than the given
      `requested_dict` must contain just 1 series! That series its name and index are
      used to  transform (i.e., **replace**) that **series with the numpy array**.
      When a user does not want a numpy array to replace its input series, it is his /
      her responsibility to create a new `pd.Series` (or `pd.DataFrame`) of that numpy
      array with a different (column) name.
    * If `func_output` is a `pd.Series`, keep in mind that the input series gets
      transformed (i.e., **replaced**) with the `func_output` **when the series name is
      equal** or when there is no series name (see note above).

    """
    if isinstance(func_output, pd.DataFrame):
        # Nothing has to be done! A pd.DataFrame can be added to a series_dict using
        # series_dict.update(df)
        # Note: converting this to a dictionary (to_dict()) is **very** inefficient!
        # Assert that the DataFrame has a time-index
        if len(func_output):
            assert isinstance(func_output.index, pd.DatetimeIndex)
        # Assert that the DataFrame columns are named
        assert all(
            func_output.columns.values != [i for i in range(func_output.shape[1])]
        )
        return func_output

    elif isinstance(func_output, pd.Series):
        # Convert series to series_dict and return
        # => if func_output.name is in the required_dict, than the original series will
        #    be replaced by this new series.
        # Assert that the series has a time-index
        assert isinstance(func_output.index, pd.DatetimeIndex)
        # Assert (func_output.name is not None) | (len(required_dict) == 1)
        if func_output.name is None:
            # If a series without a name is returned that is constructed from just 1
            # series => the input series will be replaced by this series
            assert len(required_dict) == 1
            # Rename, but keep the same reference to the data
            input_series = list(required_dict.values())[0]
            func_output.rename(input_series.name, inplace=True, copy=False)
        return {str(func_output.name): func_output}

    elif isinstance(func_output, np.ndarray):
        # Must be constructed from just 1 series
        # => the input series will be replaced by this array
        assert len(required_dict) == 1
        # Create a series with the same view
        input_series = list(required_dict.values())[0]
        return {str(input_series.name): _np_array_to_series(func_output, input_series)}

    elif isinstance(func_output, list) and all([isinstance(el, pd.Series) for el in func_output]):
        # Convert the list of series into a series dict, using the same data reference
        # => if for any series in the list, series.name is in the required_dict, than
        #    the the original series will be replaced by this new series.
        # Assert that all outputs have different names
        assert len(set([s.name for s in func_output])) == len(func_output)
        assert all([isinstance(s.index, pd.DatetimeIndex) for s in func_output])
        return {s.name: s for s in func_output}

    else:
        raise TypeError(f"Function output type is invalid for processor {func_name}")



"""Contains the used variables and functions to provide logging functionality.

See Also
--------
SeriesPipeline : its `logging_file_path` of the `process` method.

"""

__author__ = "Jeroen Van Der Donckt"

import logging
import pandas as pd
import re

from ..utils.logging import logging_file_to_df

# Package specific logger
logger = logging.getLogger("feature_processing_logger")
logger.setLevel(logging.DEBUG)

# Create logger which writes WARNING messages or higher to sys.stderr
console = logging.StreamHandler()
console.setLevel(logging.WARNING)
logger.addHandler(console)


def _parse_message(message: str) -> list:
    """Parse the message of the logged info."""
    regex = r"\[(.*?)\]"
    matches = re.findall(regex, message)
    assert len(matches) == 3
    func = matches[0]
    series_names = matches[1].replace("'", "")
    duration_s = float(matches[2].rstrip(" seconds"))
    return [func, series_names, duration_s]


def _parse_logging_execution_to_df(logging_file_path: str) -> pd.DataFrame:
    """Parse the logged messages into a dataframe that contains execution info.

    Parameters
    ----------
    logging_file_path: str
        The file path where the logged messages are stored. This is the file path that
        is passed to the ``SeriesPipeline`` its process method.

    Returns
    -------
    pd.DataFrame
        A DataFrame with the processor its method, series names and calculation 
        duration.

    Note
    ----
    This function only works when the ``logging_file_path`` used in a
    ``SeriesPipeline`` its ``process`` method is passed.

    """
    df = logging_file_to_df(logging_file_path)
    df[["function", "series_names", "duration"]] = \
        list(df["message"].apply(_parse_message))
    return df.drop(columns=["name", "log_level", "message"])


def get_processor_logs(logging_file_path: str) -> pd.DataFrame:
    """Get execution (time) info for each processor of a ``SeriesPipeline``.

    Parameters
    ----------
    logging_file_path: str
        The file path where the logged messages are stored. This is the file path that
        is passed to the ``SeriesPipeline`` its ``process`` method.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing each processor its duration and its series names.
    
    """
    df = _parse_logging_execution_to_df(logging_file_path)
    df["duration"] = pd.to_timedelta(df["duration"], unit="s")
    return df



# -*- coding: utf-8 -*-
__author__ = "Jonas Van Der Donckt, Jeroen Van Der Donckt"

from sklearn.base import BaseEstimator, TransformerMixin
import pandas as pd


from typing import List, Any, Optional


def to_dataframe_operator(operator: Any, operator_cols: Optional[List[str]] = None):
    """Convert the given (transformer) operator to a DataFrameOperator.

    Parameters
    ----------
    operator: Any
        The operator that should be transformed to a `DataFrameOperator`.
    operator_cols: List[str], optional
        The list of column names on which the operator should operate, by default None.

    Returns
    -------
    A ``DataFrameOperator`` from the given operator when eligable (i.e, an operator
    that is instance of ``TransformerMixin`` and is not yet ``DataFrameOperator``),
    otherwise the operator is just returned.

    """
    if isinstance(operator, DataFrameOperator):
        return operator
    elif isinstance(operator, TransformerMixin):
        return DataFrameOperator(operator, operator_cols)
    else:
        return operator


class DataFrameOperator(BaseEstimator, TransformerMixin):
    """Dataframe operator (transformer) for using dataframes in sklearn-pipeline.

    This is a wrapper class for `sklearn.pipeline.Pipeline` which outputs a DataFrame
    instead of a numpy array. The wrapped operator applies the given operation on
    the given operator_columns (when passed to the constructor).

    Parameters
    ---------
    wrapped_operator: TransformerMixin
        Operator (transformer) that needs to be wrapped.
    operator_cols:  List[str], optional
        The list of column names on which the operator should operate, by default None.

    """

    def __init__(
        self,
        wrapped_operator: TransformerMixin,
        operator_cols: Optional[List[str]] = None,
    ):
        """Create a dataframeoperator."""
        self.wrapped_operator = wrapped_operator
        self.operator_cols = operator_cols

    def fit(self, X: pd.DataFrame, y=None):
        # TODO: document?
        assert isinstance(X, pd.DataFrame)
        operator_cols = (
            X.columns.values if self.operator_cols is None else self.operator_cols
        )
        print(self.wrapped_operator, operator_cols)
        self.wrapped_operator.fit(X[operator_cols], y)
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        # TODO: document?
        assert isinstance(X, pd.DataFrame)
        operator_cols = (
            X.columns.values if self.operator_cols is None else self.operator_cols
        )
        try:
            data = self.wrapped_operator.transform(X[operator_cols])
            result_df = pd.DataFrame(index=X.index, data=data, columns=operator_cols)
            # Add the unused columns back to the resulting dataframe
            unused_columns = list(set(X.columns).difference(operator_cols))
            result_df[unused_columns] = X[unused_columns]
            return result_df
        except:
            cols_error = list(set(operator_cols).difference(X.columns.values))
            raise KeyError(f"The DataFrame does not include the columns: {cols_error}")



"""<i><b>flex</b>ible <b>t</b>ime-<b>s</b>eries operations</i>

.. include:: ../docs/pdoc_include/root_documentation.md
"""

__docformat__ = 'numpy'
__author__ = "Jonas Van Der Donckt, Jeroen Van Der Donckt, Emiel Deprost"
__version__ = '0.1.1.7'
__pdoc__ = {
    # do not show tue utils module
    'tsflex.utils': False,
    # show the seriesprocessor & numpyfuncwrapper their call method
    'SeriesProcessor.__call__': True,
    'NumpyFuncWrapper.__call__': True,
}

__all__ = ["__version__", "__pdoc__"]



"""Utility functions for internal data (representation) operations."""

__author__ = "Jeroen Van Der Donckt, Jonas Van Der Donckt"

import itertools
from typing import Any, Dict, Iterable, Iterator, List, Union, Tuple

import pandas as pd


def series_dict_to_df(series_dict: Dict[str, pd.Series]) -> pd.DataFrame:
    """Convert the `series_dict` into a pandas DataFrame with an outer merge.

    Parameters
    ----------
    series_dict : Dict[str, pd.Series]
        The dict with the Series.

    Returns
    -------
    pd.DataFrame
        The merged pandas DataFrame

    Note
    ----
    The `series_dict` is an internal representation of the time-series data.
    In this dictionary, the key is always the accompanying series its name.
    This internal representation is constructed in the `process` method of the
    `SeriesPipeline`.

    .. todo::
        @Jeroen - check performance

    """
    # 0. Check if the series_dict has only 1 series, to create the df efficiently
    if len(series_dict) == 1:
        return pd.DataFrame(series_dict)
    # 1. Check if the time-indexes of the series are equal, to create the df efficiently
    try:
        index_info = set(
            [
                (s.index[0], s.index[-1], len(s), s.index.freq)
                for s in series_dict.values()
            ]
        )
        if len(index_info) == 1:
            # If list(index_info)[0][-1] is None => this code assumes equal index to
            # perform efficient merge, otherwise the join will be still correct, but it
            # would actually be more efficient to perform the code at (2.).
            # But this disadvantage (slower merge) does not outweigh the time-loss when
            # checking the full index.

            # When the time-indexes are the same we can create df very efficiently
            return pd.DataFrame(series_dict, copy=False)
    except IndexError:
        # We catch an indexError as we make the assumption that there is data within the
        # series -> we do not make that assumption when constructing the DataFrame the
        # slow way.
        pass
    # 2. If check failed, create the df by merging the series (the slow way)
    df = pd.DataFrame()
    for key, s in series_dict.items():
        # Check if we deal with a valid series_dict before merging on series.name
        assert key == s.name
        df = df.merge(s, left_index=True, right_index=True, how="outer", copy=False)
    return df


def to_series_list(
    data: Union[pd.Series, pd.DataFrame, List[Union[pd.Series, pd.DataFrame]]]
) -> List[pd.Series]:
    """Convert the data to a list of series.

    Parameters
    ----------
    data : Union[pd.Series, pd.DataFrame, List[Union[pd.Series, pd.DataFrame]]
        Dataframe or Series or list thereof, that should be transformed to a series
        list.

    Returns
    -------
    List[pd.Series]
        List of series containing the series present in `data`.

    """
    if not isinstance(data, list):
        data = [data]

    series_list: List[pd.Series] = []
    for s in data:
        if isinstance(s, pd.DataFrame):
            series_list += [s[c] for c in s.columns]
        elif isinstance(s, pd.Series):
            series_list.append(s)
        else:
            raise TypeError("Non pd.Series or pd.DataFrame object passed.")

    return series_list


def to_list(x: Any) -> List:
    """Convert the input to a list if necessary.

    Parameters
    ----------
    x : Any
        The input that needs to be convert to a list.
    
    Returns
    -------
    List
        A list of `x` if `x` wasn't a list yet, otherwise `x`.

    """
    if not isinstance(x, list):
        return [x]
    return x


def to_tuple(x: Any) -> Tuple[Any]:
    """Convert the input to a tuple if necessary.

    Parameters
    ----------
    x : Any
        The input that needs to be convert to a tuple.
    
    Returns
    -------
    List
        A tuple of `x` if `x` wasn't a tuple yet, otherwise `x`.

    """
    if not isinstance(x, tuple):
        return (x,)
    return x


def flatten(data: Iterable) -> Iterator:
    """Flatten the given input data to an iterator.

    Parameters
    ----------
    data : Iterable
        The iterable data that needs to be flattened.

    Returns
    -------
    Iterator
        An iterator for the flattened data.

    """
    return itertools.chain.from_iterable(data)



# -*- coding: utf-8 -*-
"""

FeatureDescriptor and MultipleFeatureDescriptors class for creating time-series
features.

"""

import itertools
from typing import Callable, List, Union, Tuple

import pandas as pd

from .function_wrapper import NumpyFuncWrapper
from ..utils.classes import FrozenClass
from ..utils.data import to_list, to_tuple
from ..utils.time import parse_time_arg


class FeatureDescriptor(FrozenClass):
    """A FeatureDescriptor object, containing all feature information.

    Parameters
    ----------
    function : Union[NumpyFuncWrapper, Callable]
        The function that calculates this feature.
        The prototype of the function should match: \n

            function(*series: np.ndarray)
                -> Union[Any, List[Any]]

    series_name : Union[str, Tuple[str]]
        The names of the series on which the feature function should be applied.
        This argument should match the `function` its input; \n
        * If `series_name` is a string (or tuple of a single string), than 
            `function` should require just one series as input.
        * If `series_name` is a tuple of strings, than `function` should
            require `len(tuple)` series as input **and in exactly the same order**
    window : Union[float, str, pd.Timedelta]
        The window size, this argument supports multiple types: \n
        * If the type is an `float`, it represents the series its window size in
            **seconds**.
        * If the window's type is a `pd.Timedelta`, the window size represents
            the window-time.
        * If a `str`, it represents a window-time-string. \n
            .. Note::
                When no time-unit is present in the string, it represents the stride
                size in **seconds**.

    stride : Union[int, str, pd.Timedelta]
        The stride of the window rolling process, supports multiple types: \n
        * If the type is `float`, it represents the stride size in **seconds**
        * If the type is `pd.Timedelta`, it represents the stride-roll timedelta.
        * If a type is `str`, it represents a stride-roll-time-string. \n
            .. Note::
                When no time-unit is present in the string, it represents the stride
                size in **seconds**.

    Notes
    -----
    * For each `function` - `input`(-series) - `window` - stride combination, one needs
      to create a distinct `FeatureDescriptor`. Hence it is more convenient to
      create a `MultipleFeatureDescriptors` when `function` - `window` - `stride`
      _combinations_ should be applied on various input-series (combinations).
    * When `function` takes multiple series (i.e., arguments) as input, these are
      joined (based on the index) before applying the function. If the indexes of
      these series are not exactly the same, it might occur that not all series have
      exactly the same length! Hence,  make sure that the `function` can deal with
      this!
    * For more information about the str-based time args, look into:
      [pandas time delta](https://pandas.pydata.org/pandas-docs/stable/user_guide/timedeltas.html#parsing){:target="_blank"}
    <br><br>
    .. todo::
        * Add documentation of how the index/slicing takes place / which
          assumptions we make.
        * Raise error function tries to change values of view due to flag


    Raises
    ------
    TypeError
        Raised when the `function` is not an instance of Callable or
        NumpyFuncWrapper.

    See Also
    --------
    StridedRolling: As the window-stride (time) conversion takes place there.

    """

    def __init__(
        self,
        function: Union[NumpyFuncWrapper, Callable],
        series_name: Union[str, Tuple[str]],
        window: Union[float, str, pd.Timedelta],
        stride: Union[float, str, pd.Timedelta],
    ):
        self.series_name: Tuple[str] = to_tuple(series_name)
        self.window: pd.Timedelta = parse_time_arg(window)
        self.stride: pd.Timedelta = parse_time_arg(stride)

        # Order of if statements is important (as NumpyFuncWrapper also is a Callable)!
        if isinstance(function, NumpyFuncWrapper):
            self.function: NumpyFuncWrapper = function
        elif isinstance(function, Callable):
            self.function: NumpyFuncWrapper = NumpyFuncWrapper(function)
        else:
            raise TypeError(
                "Expected feature function to be a `NumpyFuncWrapper` but is a"
                f" {type(function)}."
            )

        # Construct a function-string
        f_name = str(self.function)
        self._func_str: str = f"{self.__class__.__name__} - func: {f_name}"

        self._freeze()

    def get_required_series(self) -> List[str]:
        """Return all required series names for this feature descriptor.

        Return the list of series names that are required in order to execute the
        feature function.

        Returns
        -------
        List[str]
            List of all the required series names.

        """
        return list(set(self.series_name))        

    def __repr__(self) -> str:
        """Representation string of Feature."""
        return f"{self.__class__.__name__}({self.series_name}, {self.window}, " \
               f"{self.stride})"


class MultipleFeatureDescriptors:
    """Create a MultipleFeatureDescriptors object.

    Create a list of features from **all** combinations of the given parameter
    lists. Total number of created Features will be:

        len(func_inputs)*len(functions)*len(windows)*len(strides).

    Parameters
    ----------
    functions : Union[NumpyFuncWrapper, Callable, List[Union[NumpyFuncWrapper, Callable]]]
        The functions, can be either of both types (even in a single array).
    series_names : Union[str, Tuple[str], List[str], List[Tuple[str]]]
        The names of the series on which the feature function should be applied.

        This argument should match the `function` its input; \n
        * If `series_names` is a (list of) string (or tuple of a single string),
          than `function` should require just one series as input.
        * If `series_names` is a (list of) tuple of strings, than `function` should
          require `len(tuple)` series as input.

        A list means multiple series (combinations) to extract feature from; \n
        * If `series_names` is a string or a tuple of strings, than `function` will
          be called only once for the series of this argument.
        * If `series_names` is a list of either strings or tuple of strings, than
          `function` will be called for each entry of this list.

        Note: when passing a list as `series_names`, all items in this list should
        have the same type, i.e, either \n
        * all a str
        * or, all a tuple _with same length_. \n
    windows : Union[float, str, pd.Timedelta, List[Union[float, str, pd.Timedelta]]],
        All the window sizes.
    strides : Union[float, str, pd.Timedelta, List[Union[float, str, pd.Timedelta]]],
        All the strides.

    """
    def __init__(
        self,
        functions: Union[NumpyFuncWrapper, Callable, List[Union[NumpyFuncWrapper, Callable]]],
        series_names: Union[str, Tuple[str], List[str], List[Tuple[str]]],
        windows: Union[float, str, pd.Timedelta, List[Union[float, str, pd.Timedelta]]],
        strides: Union[float, str, pd.Timedelta, List[Union[float, str, pd.Timedelta]]],
    ):
        # Cast functions to NumpyFuncWrapper, this avoids creating multiple
        # NumpyFuncWrapper objects for the same function in the FeatureDescriptor
        def to_np_func_wrapper(f: Callable): 
            return f if isinstance(f, NumpyFuncWrapper) else NumpyFuncWrapper(f)
        functions = [to_np_func_wrapper(f) for f in to_list(functions)]
        # Convert the series names to list of tuples
        series_names = [to_tuple(names) for names in to_list(series_names)]
        # Assert that function inputs (series) all have the same length
        assert all(
            len(series_names[0]) == len(series_name_tuple)
            for series_name_tuple in series_names
        )
        # Convert the other types to list
        windows = to_list(windows)
        strides = to_list(strides)

        self.feature_descriptions: List[FeatureDescriptor] = []
        # Iterate over all combinations
        combinations = [functions, series_names, windows, strides]
        for function, series_name, window, stride in itertools.product(*combinations):
            self.feature_descriptions.append(
                FeatureDescriptor(function, series_name, window, stride)
            )



"""SKFeatureCollection class for wrapped sklearn-compatible `FeatureCollection`.

.. todo::
    * **BaseEstimator support**: it is not really useful right now to support this.
    As for example sklearn GridSearchCV requires X and y to have the same length,
    but `FeatureCollection`.`calculate` (almost always) transforms the length of X in
    your pipeline.<br>
    _Possible solution_: look into sklearn-contrib imblearn how they handle this

See Also
--------
Example notebooks.

"""

from __future__ import annotations  # Make typing work for the enclosing class

__author__ = "Jeroen Van Der Donckt"

import pandas as pd

from typing import List, Optional, Union, Any
from pathlib import Path
from sklearn.base import TransformerMixin

from .feature_collection import FeatureCollection
from .feature import FeatureDescriptor, MultipleFeatureDescriptors


class SKFeatureCollection(TransformerMixin):
    """Sklearn-compatible transformer for extracting features using `FeatureCollection`.

    This class is basically a wrapper around `FeatureCollection` and its `calculate`
    method, enabling sklearn compatibility, e.g. including an instance of this class
    in `sklearn.pipeline.Pipeline`. The constructor wraps the `FeatureCollection` its
    constructor arguments and the  relevant parameters of its `calculate` method.

    Parameters
    ----------
    feature_descriptors : Union[FeatureDescriptor, MultipleFeatureDescriptors, List[Union[FeatureDescriptor, MultipleFeatureDescriptors]]]
        Features to include in the collection.
    logging_file_path : Union[str, Path], optional
        The file path where the logged messages are stored. If `None`, then no
        logging `FileHandler` will be used and the logging messages are only pushed
        to stdout. Otherwise, a logging `FileHandler` will write the logged messages
        to the given file path.
    n_jobs : int, optional
        The number of processes used for the feature calculation. If `None`, then
        the number returned by `os.cpu_count()` is used, by default None. \n
        If n_jobs is either 0 or 1, the code will be executed sequentially without
        creating a process pool. This is very useful when debugging, as the stack
        trace will be more comprehensible.

    Notes
    -----
    * If a `logging_file_path` is provided, the execution (time) info can be
      retrieved by calling `logger.get_feature_logs(logging_file_path)`. <br>
      Be aware that the `logging_file_path` gets cleared before the logger pushes
      logged messages. Hence, one should use a separate logging file for each
      constructed processing and feature instance with this library.
    * It is not possible to pass the `merge_dfs` argument from the `calculate`
      method, because this should always be True (in order to output a valid
      iterable in the `transform` method)\n
    The following three changes were necessary to enable this sklearn-compatibility: \n
    1. The parameters of the `FeatureCollection` its constructor are logged in this
       class.
    2. The relevant parameters of the `FeatureCollection` its `calculate` method have
       been moved to the constructor and are also logged.
    3. The `FeatureCollection` its `calculate` method is wrapped in this class its
       `transform` method,

    """

    def __init__(
        self,
        feature_descriptors: Union[FeatureDescriptor, MultipleFeatureDescriptors, List[
            Union[FeatureDescriptor, MultipleFeatureDescriptors]]],
        logging_file_path: Optional[Union[str, Path]] = None,
        n_jobs: Optional[int] = None,
    ):
        self.feature_descriptors: Any = feature_descriptors
        self.logging_file_path = logging_file_path
        self.n_jobs = n_jobs

    def fit(self, X=None, y=None) -> SKFeatureCollection:
        """Fit function that is not needed for this transformer.

        Note
        ----
        This function does nothing and is here for compatibility reasons.

        Parameters
        ----------
        X : Any
            Unneeded.
        y : Any
            Unneeded.

        Returns
        -------
        SKFeatureCollection
            The transformer instance itself is returned.

        """
        return self

    def transform(
        self,
        X: Union[pd.Series, pd.DataFrame, List[Union[pd.Series, pd.DataFrame]]],
    ) -> pd.DataFrame:
        """Calculate features on the data.

        Parameters
        ----------
        X : Union[pd.Series, pd.DataFrame, List[Union[pd.Series, pd.DataFrame]]]
            Dataframe or Series or list thereof, with all the required data for the
            feature calculation.

            **Remark**: each Series/DataFrame must have a `pd.DatetimeIndex`. \n
            Also note that this parameter corresponds to the `data` parameter of the
            `FeatureCollection` its `calculate` method.

        Returns
        -------
        pd.DataFrame
            A DataFrame, containing the calculated features.

        """
        feature_collection = FeatureCollection(self.feature_descriptors)
        return feature_collection.calculate(
            data=X,
            return_df=True,
            show_progress=False,
            logging_file_path=self.logging_file_path,
            n_jobs=self.n_jobs,
        )

    def __repr__(self) -> str:
        """Representation string of a SKFeatureCollection."""
        return repr(FeatureCollection(self.feature_descriptors))



# -*- coding: utf-8 -*- # TODO: zetten we dat nu overal of niet?
"""Object-oriented utilities."""

__author__ = 'Jonas Van Der Donckt'


class FrozenClass(object):
    """Superclass which allows subclasses to freeze at any time."""

    __is_frozen = False

    def __setattr__(self, key, value):
        if self.__is_frozen and not hasattr(self, key):
            raise TypeError("%r is a frozen class" % self)
        object.__setattr__(self, key, value)

    def _freeze(self):
        self.__is_frozen = True



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tsflex',
 'tsflex.chunking',
 'tsflex.features',
 'tsflex.pipeline_sklearn_dev',
 'tsflex.processing',
 'tsflex.utils']

package_data = \
{'': ['*']}

install_requires = \
['dill>=0.3.3,<0.4.0',
 'fastparquet>=0.6.3,<0.7.0',
 'numpy>=1.19.0,<2.0.0',
 'pandas>=1.2.3,<2.0.0',
 'pathos>=0.2.7,<0.3.0',
 'tqdm>=4.60.0,<5.0.0']

setup_kwargs = {
    'name': 'tsflex',
    'version': '0.1.1.7',
    'description': 'Toolkit for flexible processing & feature extraction on time-series data',
    'long_description': '# <p align="center"> <a href="https://predict-idlab.github.io/tsflex"><img alt="tsflex" src="https://raw.githubusercontent.com/predict-idlab/tsflex/main/docs/_static/logo.png" height="100"></a></p>\n\n[![PyPI Latest Release](https://img.shields.io/pypi/v/tsflex.svg)](https://pypi.org/project/tsflex/)\n[![support-version](https://img.shields.io/pypi/pyversions/tsflex)](https://img.shields.io/pypi/pyversions/tsflex)\n[![codecov](https://codecov.io/gh/predict-idlab/tsflex/branch/main/graph/badge.svg)](https://codecov.io/gh/predict-idlab/tsflex)\n[![Code quality](https://img.shields.io/lgtm/grade/python/g/predict-idlab/tsflex.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/predict-idlab/tsflex/context:python)\n![License: MIT](https://img.shields.io/badge/License-MIT-blue.svg)\n[![Downloads](https://pepy.tech/badge/tsflex)](https://pepy.tech/project/tsflex)\n[![Documentation](https://github.com/predict-idlab/tsflex/actions/workflows/deploy-docs.yml/badge.svg)](https://github.com/predict-idlab/tsflex/actions/workflows/deploy-docs.yml)\n[![Testing](https://github.com/predict-idlab/tsflex/actions/workflows/test.yml/badge.svg)](https://github.com/predict-idlab/tsflex/actions/workflows/test.yml)\n\n*tsflex* is a toolkit for _**flex**ible **t**ime-**s**eries_ **[processing](https://predict-idlab.github.io/tsflex/processing) & [feature extraction](https://predict-idlab.github.io/tsflex/features)**, making few assumptions about input data. \n\n#### Useful links\n\n- [Documentation](https://predict-idlab.github.io/tsflex/)\n- [Example notebooks](https://github.com/predict-idlab/tsflex/tree/main/examples)\n\n## Installation\n\nIf you are using [**pip**](https://pypi.org/project/tsflex/), just execute the following command:\n\n```sh\npip install tsflex\n```\n\n## Why tsflex? ✨\n\n* flexible;\n    * handles multi-variate time-series\n    * versatile function support  \n      => **integrates natively** with many packages for processing (e.g., [scipy.signal](https://docs.scipy.org/doc/scipy/reference/tutorial/signal.html)) & feature extraction (e.g., [numpy](https://numpy.org/doc/stable/reference/routines.html), [scipy.stats](https://docs.scipy.org/doc/scipy/reference/tutorial/stats.html))\n    * feature-extraction handles **multiple strides & window sizes**\n* efficient view-based operations  \n  => extremely **low memory peak & fast execution** times ([see benchmarks]())\n    <!-- * faster than any existing library (single- & multi-core)\n    * lower memory peak than any existing library (single- & multi-core) -->\n* maintains the **time-index** of the data\n* makes **little to no assumptions** about the time-series data\n\n## Usage\n\n_tsflex_ is built to be intuitive, so we encourage you to copy-paste this code and toy with some parameters!\n\n### <a href="https://predict-idlab.github.io/tsflex/features/#getting-started">Feature extraction</a>\n\n```python\nimport pandas as pd; import scipy.stats as ssig; import numpy as np\nfrom tsflex.features import FeatureDescriptor, FeatureCollection, NumpyFuncWrapper\n\n# 1. -------- Get your time-indexed data --------\n# Data contains 1 column; ["TMP"]\nurl = "https://github.com/predict-idlab/tsflex/raw/main/examples/data/empatica/tmp.parquet"\ndata = pd.read_parquet(url).set_index("timestamp")\n\n# 2 -------- Construct your feature collection --------\nfc = FeatureCollection(\n    feature_descriptors=[\n        FeatureDescriptor(\n            function=NumpyFuncWrapper(func=ssig.skew, output_names="skew"),\n            series_name="TMP", \n            window="5min",  # Use 5 minutes \n            stride="2.5min",  # With steps of 2.5 minutes\n        )\n    ]\n)\n# -- 2.1. Add features to your feature collection\nfc.add(FeatureDescriptor(np.min, "TMP", \'2.5min\', \'2.5min\'))\n\n# 3 -------- Calculate features --------\nfc.calculate(data=data)\n```\n\n### More examples\n\nFor processing [look here](https://predict-idlab.github.io/tsflex/processing/index.html#working-example)    \nOther examples can be found [here](https://github.com/predict-idlab/tsflex/tree/main/examples)\n\n## Future work 🔨\n\n* scikit-learn integration for both processing and feature extraction<br>\n  **note**: is actively developed upon [sklearn integration](https://github.com/predict-idlab/tsflex/tree/sklearn_integration) branch.\n* support for multi-indexed dataframes\n* random-strided rolling for data-augmention purposes.\n\n## Referencing our package\n\nIf you use `tsflex` in a scientific publication, we would highly appreciate citing us as:\n\n```bibtex\n@article{vanderdonckt2021tsflex,\n    author = {Van Der Donckt, Jonas and Van Der Donckt, Jeroen and Van Hoecke, Sofie},\n    title = {tsflex: flexible time series processing \\& feature extraction},\n    journal = {SoftwareX},\n    year = {2021},\n    url = {https://github.com/predict-idlab/tsflex},\n    publisher={Elsevier}\n}\n```\n\n---\n\n<p align="center">\n👤 <i>Jonas Van Der Donckt, Jeroen Van Der Donckt, Emiel Deprost</i>\n</p>\n',
    'author': 'Jonas Van Der Donckt, Jeroen Van Der Donckt, Emiel Deprost',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/predict-idlab/tsflex',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



"""Processing module.

.. include:: ../../docs/pdoc_include/processing.md

"""

__author__ = "Jonas Van Der Donckt, Emiel Deprost, Jeroen Van Der Donckt"

from .logger import get_processor_logs
from .series_pipeline import SeriesPipeline
from .series_processor import SeriesProcessor, dataframe_func
from .. import __pdoc__

__pdoc__['SeriesProcessor.__call__'] = True

__all__ = [
    "dataframe_func",
    "SeriesProcessor",
    "SeriesPipeline",
    "get_processor_logs",
]



"""SKSeriesPipeline class for wrapped sklearn-compatible `SeriesPipeline`.

See Also
--------
Example notebooks.

"""

from __future__ import annotations  # Make typing work for the enclosing class

__author__ = "Jeroen Van Der Donckt"

import pandas as pd

from typing import List, Optional, Union
from pathlib import Path
from sklearn.base import TransformerMixin

from .series_processor import SeriesProcessor
from .series_pipeline import SeriesPipeline


# --- Future work ---
# * BaseEstimator support: it is not really useful right now to support this.
#   As for example sklearn GridSearchCV requires X and y to have the same length,
#   but SeriesPipeline (sometimes) transforms the length of X in your pipeline.
#   => Possible solution; look into sklearn-contrib imblearn how they handle this


class SKSeriesPipeline(TransformerMixin):
    """Sklearn-compatible transformer for processing data using a `SeriesPipeline`.

    This class is basically a wrapper around `SeriesPipeline` and its `process`
    method, enabling sklearn compatibility, e.g. including an instance of this class
    in `sklearn.pipeline.Pipeline`.

    Concretely three changes were necessary to enable this sklearn-compatibility

    1. The parameters of the `SeriesPipeline` its constructor are logged in this
       class.
    2. The relevant parameters of the `SeriesPipeline` its `process` method have
       been moved to the constructor and are also logged.
    3. The `SeriesPipeline` its `process` method is wrapped in this class its
       `transform` method,

    This constructor wraps the `SeriesPipeline` constructor arguments and the
    relevant parameters of its `process` method.

    Parameters
    ----------
    processors : List[SeriesProcessor]
        List of `SeriesProcessor` objects that will be applied sequentially to the
        given data. The processing steps will be executed in the  same order as
        passed with this list.
    return_all_series : bool, optional
        Whether the output needs to return all the series, by default True. \n
        * If `True` the output will contain all series that were passed to this
          method.
        * If `False` the output will contain just the required series (see
          `SeriesPipeline.get_required_series`).
    drop_keys : List[str], optional
        Which keys should be dropped when returning the output, by default None.
    logging_file_path : Union[str, Path], optional
        The file path where the logged messages are stored, by default None. \n
        * If `None`, then no logging `FileHandler` will be used and the logging
          messages are only pushed to stdout.
        * Otherwise, a logging `FileHandler` will write the logged messages to
          the given file path.

    Notes
    -----
    * If a `logging_file_path` is provided, the execution (time) statistics can be
      retrieved by calling `logger.get_processor_logs(logging_file_path)`. <br>
      Be aware that the `logging_file_path` gets cleared before the logger pushes
      logged messages. Hence, one should use a separate logging file for each
      constructed processing and feature instance with this library.
    * It is not possible to pass the `return_df` argument from the `process`
      method, because this should always be True (in order to output a valid
      iterable in the `transform` method)
    """

    def __init__(
        self,
        processors: List[SeriesProcessor],
        return_all_series: Optional[bool] = True,
        drop_keys: Optional[List[str]] = None,
        logging_file_path: Optional[Union[str, Path]] = None,
    ):
        self.processors = processors
        self.return_all_series = return_all_series
        self.drop_keys = drop_keys
        self.logging_file_path = logging_file_path

    def fit(self, X=None, y=None) -> SKSeriesPipeline:
        """Fit function that is not needed for this transformer.

        Note
        ----
        This function does nothing and is here for compatibility reasons.

        Parameters
        ----------
        X : Any
            Unneeded.
        y : Any
            Unneeded.

        Returns
        -------
        SKSeriesPipeline
            The transformer instance itself is returned.
        """
        return self

    def transform(
        self,
        X: Union[pd.Series, pd.DataFrame, List[Union[pd.Series, pd.DataFrame]]]
    ) -> pd.DataFrame:
        """Process the data.

        Parameters
        ----------
        X : Union[pd.Series, pd.DataFrame, List[Union[pd.Series, pd.DataFrame]]]
            Dataframe or Series or list thereof, with all the required data for the
            processing steps. \n
            **Remark**: each Series/DataFrame must have a `pd.DatetimeIndex`. \n
            Note that this parameter corresponds to the `data` parameter of the
            `SeriesPipeline` its `process` method.

        Returns
        -------
        pd.DataFrame
            A DataFrame, containing the processed data.

        """
        series_pipeline = SeriesPipeline(self.processors)
        return series_pipeline.process(
            data=X,
            return_all_series=self.return_all_series,
            return_df=True,
            drop_keys=self.drop_keys,
            logging_file_path=self.logging_file_path,
        )

    def __repr__(self) -> str:
        """Representation string of a SKSeriesPipeline."""
        return repr(SeriesPipeline(self.processors))



"""Utility functions for time-related operations."""

__author__ = 'Jonas Van Der Donckt'

import pandas as pd
from typing import Union


def timedelta_to_str(td: pd.Timedelta) -> str:
    """Construct a tight string representation for the given timedelta arg.

    Parameters
    ----------
    td: pd.Timedelta
        The timedelta for which the string representation is constructed

    Returns
    -------
    str:
        The tight string bounds of format '$d-$h$m$s.$ms'.

    """
    out_str = ""

    # Edge case if we deal with negative
    if td < pd.Timedelta(seconds=0):
        td *= -1
        out_str += 'NEG'

    # Note: this must happen after the *= -1
    c = td.components
    if c.days > 0:
        out_str += f'{c.days}D'
    if c.hours > 0 or c.minutes > 0 or c.seconds > 0 or c.milliseconds > 0:
        out_str += '_' if len(out_str) else ""

    if c.hours > 0:
        out_str += f"{c.hours}h"
    if c.minutes > 0:
        out_str += f"{c.minutes}m"
    if c.seconds > 0 or c.milliseconds > 0:
        out_str += f"{c.seconds}"
        if c.milliseconds:
            out_str += f".{str(c.milliseconds / 1000).split('.')[-1].rstrip('0')}"
        out_str += "s"
    return out_str


def parse_time_arg(arg: Union[float, str, pd.Timedelta]) -> pd.Timedelta:
    """Parse the `window`/`stride` arg into a fixed set of types.

    Parameters
    ----------
    arg : Union[float, str, pd.Timedelta]
        The arg that will be parsed. \n
        * If the type is an `int` or `float`, it should represent the timedelta in
          **seconds**.
        * If the type is a `pd.Timedelta`, nothing will happen.
        * If the type is a `str`, `arg` should represent a time-string, and will be
          converted to a `pd.Timedelta`.
          Note: if there is no time-unit in the string, it should represent the
          timedelta in **seconds**.

    Returns
    -------
    pd.Timedelta
        The parsed time arg

    Raises
    ------
    TypeError
        Raised when `arg` is not an instance of `float`, `int`, `str`, or
        `pd.Timedelta`.

    """
    try:
        # Cast a string without time units to float (or cast an int to float)
        arg = float(arg)
    except:
        pass

    if isinstance(arg, float):
        return pd.Timedelta(seconds=arg)
    elif isinstance(arg, str):
        return pd.Timedelta(arg)
    raise TypeError(f"arg type {type(arg)} is not supported!")